""" Convert ARDCollections to QGIS Layer format"""


from collections import defaultdict
from uuid import uuid4

from pyproj import Proj

from max_ard.dependency_support import HAS_FIONA
from max_ard.exceptions import MissingDependency

if HAS_FIONA:
    import fiona

__all__ = ("QlrDoc",)


def QlrDoc(collection, public=False):
    """Return a an ARDCollection as a QGIS layer file

    Parameters
    ----------
    collection : ARDCollection
    public : bool, optional
      True if the data is in a public S3 bucket location (see Notes)

    Returns
    -------
    str
      QLR document contents of the ARDCollection

    Notes
    -----
    Passing `public=true` will convert s3:// URLs to the public http:// S3 endpoints.
    This functionality does not work yet for other storage types. It's functionality
    is primarily for creating QLRs of the official Maxar ARD sample datasets at
    http://maxar-ard-samples.s3-website-us-east-1.amazonaws.com/
    """

    return QLayer(collection, public).as_text()


def vsify(tile, asset, public):
    if public:
        if "s3" in tile.gdal_prefix:
            bucket, key = tile.asset_paths[asset].split("/", 1)
            return f"/vsicurl/https://{bucket}.s3.amazonaws.com/{key}"
        else:
            raise ValueError("Public files in Azure and Google Cloud not supported yet")
    else:
        return f"{tile.gdal_prefix}{tile.asset_paths[asset]}"


class ItemLayer:
    def __init__(self, tile, uuid, public=False):
        self.tile = tile
        self.name = f"{self.tile.acq_id} STAC Item"
        self.uuid = uuid
        self.proj = Proj(tile.cell.proj)
        self.public = public

    def as_treelayer(self):

        # source="{self.tile.gdal_prefix}{self.tile.asset_paths['item']}|layername={self.tile.acq_id}">
        return f"""<layer-tree-layer checked="Qt::Checked" 
            id="{self.tile.acq_id}_json_{self.uuid}" 
            expanded="1" name="{self.name}" 
            legend_exp="" providerKey="ogr" 
            source="{vsify(self.tile, 'item', self.public)}|layername={self.tile.acq_id}">
        <customproperties/>
        </layer-tree-layer>"""

    def as_maplayer(self):
        # <datasource>{self.tile.gdal_prefix}{self.tile.asset_paths['item']}|layername={self.tile.acq_id}</datasource>

        return f"""

    <maplayer geometry="Polygon" simplifyLocal="1" autoRefreshEnabled="0" minScale="100000000" refreshOnNotifyEnabled="0" simplifyDrawingTol="1" refreshOnNotifyMessage="" maxScale="0" type="vector" hasScaleBasedVisibilityFlag="0" simplifyDrawingHints="1" readOnly="0" simplifyAlgorithm="0" styleCategories="AllStyleCategories" simplifyMaxScale="1" autoRefreshTime="0" labelsEnabled="0" wkbType="Polygon">
      <id>{self.tile.acq_id}_json_{self.uuid}</id>
      <datasource>{vsify(self.tile, 'item', self.public)}|layername={self.tile.acq_id}</datasource>
      <keywordList>
        <value></value>
      </keywordList>
      <layername>{self.name}</layername>
       <srs>
        <spatialrefsys>
          <wkt>GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4326"]]</wkt>
          <proj4>+proj=longlat +datum=WGS84 +no_defs</proj4>
          <srsid>3452</srsid>
          <srid>4326</srid>
          <authid>EPSG:4326</authid>
          <description>WGS 84</description>
          <projectionacronym>longlat</projectionacronym>
          <ellipsoidacronym>WGS84</ellipsoidacronym>
          <geographicflag>true</geographicflag>
        </spatialrefsys>
      </srs>
      <provider encoding="UTF-8">ogr</provider>
      <legend type="default-vector"/>
      <map-layer-style-manager current="default">
        <map-layer-style name="default"/>
      </map-layer-style-manager>
      <flags>
        <Identifiable>1</Identifiable>
        <Removable>1</Removable>
        <Searchable>1</Searchable>
      </flags>
      <renderer-v2 forceraster="0" symbollevels="0" enableorderby="0" type="singleSymbol">
        <symbols>
          <symbol name="0" alpha="1" force_rhr="0" clip_to_extent="1" type="fill">
            <layer enabled="1" class="SimpleFill" pass="0" locked="0">
              <prop v="3x:0,0,0,0,0,0" k="border_width_map_unit_scale"/>
              <prop v="164,113,88,255" k="color"/>
              <prop v="bevel" k="joinstyle"/>
              <prop v="0,0" k="offset"/>
              <prop v="3x:0,0,0,0,0,0" k="offset_map_unit_scale"/>
              <prop v="MM" k="offset_unit"/>
              <prop v="35,35,35,255" k="outline_color"/>
              <prop v="solid" k="outline_style"/>
              <prop v="0.26" k="outline_width"/>
              <prop v="MM" k="outline_width_unit"/>
              <prop v="no" k="style"/>
              <data_defined_properties>
                <Option type="Map">
                  <Option name="name" type="QString" value=""/>
                  <Option name="properties"/>
                  <Option name="type" type="QString" value="collection"/>
                </Option>
              </data_defined_properties>
            </layer>
          </symbol>
        </symbols>
        <rotation/>
        <sizescale/>
      </renderer-v2>
      <labeling type="simple">
        <settings calloutType="simple">
          <text-style fontSize="10" textOrientation="horizontal" fontKerning="1" isExpression="1" useSubstitutions="0" previewBkgrdColor="255,255,255,255" fontLetterSpacing="0" textOpacity="1" multilineHeight="1" fontFamily=".SF NS Text" fontStrikeout="0" fontCapitals="0" fontWeight="50" fontSizeMapUnitScale="3x:0,0,0,0,0,0" fontSizeUnit="Point" fieldName="&quot;catalog_id&quot;  +  '\n'  + &#xa; format_date( &quot;datetime&quot;, 'yyyy-MM-dd' ) +  '\n'  + &#xa; &quot;quadkey&quot;  +  '\n'  + &#xa; 'Platform: '  +  &quot;platform&quot;  +  '\n'  + &#xa; 'GSD: '  +  to_string(&quot;gsd_avg&quot;)  +  '\n'    + &#xa; 'Clouds: ' +  to_string(&quot;tile:clouds_percent&quot;)&#xa; " fontWordSpacing="0" namedStyle="Regular" textColor="0,0,0,255" fontItalic="0" fontUnderline="0" blendMode="0">
            <text-buffer bufferSizeMapUnitScale="3x:0,0,0,0,0,0" bufferJoinStyle="128" bufferSizeUnits="MM" bufferOpacity="1" bufferBlendMode="0" bufferNoFill="1" bufferSize="0.9" bufferDraw="1" bufferColor="255,255,255,255"/>
            <text-mask maskOpacity="1" maskEnabled="0" maskSizeUnits="MM" maskJoinStyle="128" maskedSymbolLayers="" maskType="0" maskSizeMapUnitScale="3x:0,0,0,0,0,0" maskSize="1.5"/>
            <background shapeRotationType="0" shapeRadiiMapUnitScale="3x:0,0,0,0,0,0" shapeSizeUnit="MM" shapeRadiiUnit="MM" shapeSVGFile="" shapeOpacity="1" shapeSizeType="0" shapeOffsetMapUnitScale="3x:0,0,0,0,0,0" shapeJoinStyle="64" shapeRotation="0" shapeBorderWidth="0" shapeBorderWidthMapUnitScale="3x:0,0,0,0,0,0" shapeOffsetY="0" shapeFillColor="255,255,255,255" shapeRadiiX="0" shapeOffsetX="0" shapeBlendMode="0" shapeBorderColor="128,128,128,255" shapeType="0" shapeSizeY="0.7" shapeBorderWidthUnit="MM" shapeOffsetUnit="MM" shapeRadiiY="0" shapeDraw="1" shapeSizeMapUnitScale="3x:0,0,0,0,0,0" shapeSizeX="0.7">
              <symbol name="markerSymbol" alpha="1" force_rhr="0" clip_to_extent="1" type="marker">
                <layer enabled="1" class="SimpleMarker" pass="0" locked="0">
                  <prop v="0" k="angle"/>
                  <prop v="114,155,111,255" k="color"/>
                  <prop v="1" k="horizontal_anchor_point"/>
                  <prop v="bevel" k="joinstyle"/>
                  <prop v="circle" k="name"/>
                  <prop v="0,0" k="offset"/>
                  <prop v="3x:0,0,0,0,0,0" k="offset_map_unit_scale"/>
                  <prop v="MM" k="offset_unit"/>
                  <prop v="35,35,35,255" k="outline_color"/>
                  <prop v="solid" k="outline_style"/>
                  <prop v="0" k="outline_width"/>
                  <prop v="3x:0,0,0,0,0,0" k="outline_width_map_unit_scale"/>
                  <prop v="MM" k="outline_width_unit"/>
                  <prop v="diameter" k="scale_method"/>
                  <prop v="2" k="size"/>
                  <prop v="3x:0,0,0,0,0,0" k="size_map_unit_scale"/>
                  <prop v="MM" k="size_unit"/>
                  <prop v="1" k="vertical_anchor_point"/>
                  <data_defined_properties>
                    <Option type="Map">
                      <Option name="name" type="QString" value=""/>
                      <Option name="properties"/>
                      <Option name="type" type="QString" value="collection"/>
                    </Option>
                  </data_defined_properties>
                </layer>
              </symbol>
            </background>
            <shadow shadowOffsetAngle="135" shadowDraw="1" shadowRadius="1.5" shadowOffsetMapUnitScale="3x:0,0,0,0,0,0" shadowUnder="0" shadowColor="0,0,0,255" shadowOffsetUnit="MM" shadowRadiusAlphaOnly="0" shadowScale="100" shadowOffsetDist="1.2" shadowOffsetGlobal="1" shadowRadiusUnit="MM" shadowRadiusMapUnitScale="3x:0,0,0,0,0,0" shadowOpacity="0.7" shadowBlendMode="6"/>
            <dd_properties>
              <Option type="Map">
                <Option name="name" type="QString" value=""/>
                <Option name="properties"/>
                <Option name="type" type="QString" value="collection"/>
              </Option>
            </dd_properties>
            <substitutions/>
          </text-style>
          <text-format useMaxLineLengthForAutoWrap="1" leftDirectionSymbol="&lt;" reverseDirectionSymbol="0" plussign="0" rightDirectionSymbol=">" decimals="3" wrapChar="" addDirectionSymbol="0" autoWrapLength="0" multilineAlign="0" placeDirectionSymbol="0" formatNumbers="0"/>
          <placement centroidInside="1" overrunDistanceMapUnitScale="3x:0,0,0,0,0,0" centroidWhole="0" overrunDistance="0" preserveRotation="1" dist="0" geometryGenerator="" maxCurvedCharAngleIn="25" repeatDistance="0" geometryGeneratorType="PointGeometry" offsetUnits="MM" repeatDistanceMapUnitScale="3x:0,0,0,0,0,0" xOffset="0" priority="5" geometryGeneratorEnabled="0" offsetType="0" maxCurvedCharAngleOut="-25" yOffset="0" repeatDistanceUnits="MM" overrunDistanceUnit="MM" quadOffset="4" labelOffsetMapUnitScale="3x:0,0,0,0,0,0" distMapUnitScale="3x:0,0,0,0,0,0" predefinedPositionOrder="TR,TL,BR,BL,R,L,TSR,BSR" placementFlags="10" distUnits="MM" fitInPolygonOnly="1" layerType="PolygonGeometry" rotationAngle="0" placement="0"/>
          <rendering obstacle="1" upsidedownLabels="0" limitNumLabels="0" obstacleType="1" displayAll="0" maxNumLabels="2000" drawLabels="1" fontMinPixelSize="3" fontMaxPixelSize="10000" obstacleFactor="1" mergeLines="0" scaleMin="0" scaleVisibility="0" labelPerPart="0" minFeatureSize="0" scaleMax="0" fontLimitPixelSize="0" zIndex="0"/>
          <dd_properties>
            <Option type="Map">
              <Option name="name" type="QString" value=""/>
              <Option name="properties"/>
              <Option name="type" type="QString" value="collection"/>
            </Option>
          </dd_properties>
          <callout type="simple">
            <Option type="Map">
              <Option name="anchorPoint" type="QString" value="pole_of_inaccessibility"/>
              <Option name="ddProperties" type="Map">
                <Option name="name" type="QString" value=""/>
                <Option name="properties"/>
                <Option name="type" type="QString" value="collection"/>
              </Option>
              <Option name="drawToAllParts" type="bool" value="false"/>
              <Option name="enabled" type="QString" value="0"/>
              <Option name="lineSymbol" type="QString" value="&lt;symbol name=&quot;symbol&quot; alpha=&quot;1&quot; force_rhr=&quot;0&quot; clip_to_extent=&quot;1&quot; type=&quot;line&quot;>&lt;layer enabled=&quot;1&quot; class=&quot;SimpleLine&quot; pass=&quot;0&quot; locked=&quot;0&quot;>&lt;prop v=&quot;square&quot; k=&quot;capstyle&quot;/>&lt;prop v=&quot;5;2&quot; k=&quot;customdash&quot;/>&lt;prop v=&quot;3x:0,0,0,0,0,0&quot; k=&quot;customdash_map_unit_scale&quot;/>&lt;prop v=&quot;MM&quot; k=&quot;customdash_unit&quot;/>&lt;prop v=&quot;0&quot; k=&quot;draw_inside_polygon&quot;/>&lt;prop v=&quot;bevel&quot; k=&quot;joinstyle&quot;/>&lt;prop v=&quot;60,60,60,255&quot; k=&quot;line_color&quot;/>&lt;prop v=&quot;solid&quot; k=&quot;line_style&quot;/>&lt;prop v=&quot;0.3&quot; k=&quot;line_width&quot;/>&lt;prop v=&quot;MM&quot; k=&quot;line_width_unit&quot;/>&lt;prop v=&quot;0&quot; k=&quot;offset&quot;/>&lt;prop v=&quot;3x:0,0,0,0,0,0&quot; k=&quot;offset_map_unit_scale&quot;/>&lt;prop v=&quot;MM&quot; k=&quot;offset_unit&quot;/>&lt;prop v=&quot;0&quot; k=&quot;ring_filter&quot;/>&lt;prop v=&quot;0&quot; k=&quot;use_custom_dash&quot;/>&lt;prop v=&quot;3x:0,0,0,0,0,0&quot; k=&quot;width_map_unit_scale&quot;/>&lt;data_defined_properties>&lt;Option type=&quot;Map&quot;>&lt;Option name=&quot;name&quot; type=&quot;QString&quot; value=&quot;&quot;/>&lt;Option name=&quot;properties&quot;/>&lt;Option name=&quot;type&quot; type=&quot;QString&quot; value=&quot;collection&quot;/>&lt;/Option>&lt;/data_defined_properties>&lt;/layer>&lt;/symbol>"/>
              <Option name="minLength" type="double" value="0"/>
              <Option name="minLengthMapUnitScale" type="QString" value="3x:0,0,0,0,0,0"/>
              <Option name="minLengthUnit" type="QString" value="MM"/>
              <Option name="offsetFromAnchor" type="double" value="0"/>
              <Option name="offsetFromAnchorMapUnitScale" type="QString" value="3x:0,0,0,0,0,0"/>
              <Option name="offsetFromAnchorUnit" type="QString" value="MM"/>
              <Option name="offsetFromLabel" type="double" value="0"/>
              <Option name="offsetFromLabelMapUnitScale" type="QString" value="3x:0,0,0,0,0,0"/>
              <Option name="offsetFromLabelUnit" type="QString" value="MM"/>
            </Option>
          </callout>
        </settings>
      </labeling>
      <blendMode>0</blendMode>
      <featureBlendMode>0</featureBlendMode>
      <layerOpacity>1</layerOpacity>
      <SingleCategoryDiagramRenderer diagramType="Histogram" attributeLegend="1">
        <DiagramCategory showAxis="1" sizeType="MM" lineSizeScale="3x:0,0,0,0,0,0" labelPlacementMethod="XHeight" penColor="#000000" spacing="5" width="15" rotationOffset="270" maxScaleDenominator="1e+08" scaleDependency="Area" penWidth="0" lineSizeType="MM" minimumSize="0" barWidth="5" backgroundAlpha="255" direction="0" sizeScale="3x:0,0,0,0,0,0" penAlpha="255" minScaleDenominator="0" backgroundColor="#ffffff" enabled="0" spacingUnit="MM" spacingUnitScale="3x:0,0,0,0,0,0" opacity="1" height="15" diagramOrientation="Up" scaleBasedVisibility="0">
          <fontProperties description=".SF NS Text,13,-1,5,50,0,0,0,0,0" style=""/>
          <axisSymbol>
            <symbol name="" alpha="1" force_rhr="0" clip_to_extent="1" type="line">
              <layer enabled="1" class="SimpleLine" pass="0" locked="0">
                <prop v="square" k="capstyle"/>
                <prop v="5;2" k="customdash"/>
                <prop v="3x:0,0,0,0,0,0" k="customdash_map_unit_scale"/>
                <prop v="MM" k="customdash_unit"/>
                <prop v="0" k="draw_inside_polygon"/>
                <prop v="bevel" k="joinstyle"/>
                <prop v="35,35,35,255" k="line_color"/>
                <prop v="solid" k="line_style"/>
                <prop v="0.26" k="line_width"/>
                <prop v="MM" k="line_width_unit"/>
                <prop v="0" k="offset"/>
                <prop v="3x:0,0,0,0,0,0" k="offset_map_unit_scale"/>
                <prop v="MM" k="offset_unit"/>
                <prop v="0" k="ring_filter"/>
                <prop v="0" k="use_custom_dash"/>
                <prop v="3x:0,0,0,0,0,0" k="width_map_unit_scale"/>
                <data_defined_properties>
                  <Option type="Map">
                    <Option name="name" type="QString" value=""/>
                    <Option name="properties"/>
                    <Option name="type" type="QString" value="collection"/>
                  </Option>
                </data_defined_properties>
              </layer>
            </symbol>
          </axisSymbol>
        </DiagramCategory>
      </SingleCategoryDiagramRenderer>
      <blendMode>0</blendMode>
      <featureBlendMode>0</featureBlendMode>
      <layerOpacity>1</layerOpacity>
      <geometryOptions geometryPrecision="0" removeDuplicateNodes="0">
        <activeChecks type="StringList">
          <Option type="QString" value=""/>
        </activeChecks>
        <checkConfiguration/>
      </geometryOptions>
      <attributeactions>
        <defaultAction key="Canvas" value="{00000000-0000-0000-0000-000000000000}"/>
      </attributeactions>
    </maplayer> """


class VisualDataLayer:
    def __init__(self, tile, uuid, public):
        self.tile = tile
        self.name = f"{self.tile.acq_id} Visual Image"
        self.uuid = uuid
        self.proj = Proj(tile.cell.proj)
        self.public = public
        self.pan = tile.properties["platform"] == "WV01"

    def as_treelayer(self):

        return f"""<layer-tree-layer checked="Qt::Checked" 
            id="{self.tile.acq_id}_visual_tif_{self.uuid}" 
            expanded="1" name="{self.name}" 
            legend_exp="" providerKey="gdal" 
            source="{vsify(self.tile, 'visual', self.public)}">
        <customproperties/>
        </layer-tree-layer>"""

    def as_maplayer(self):

        if self.pan:
            nodata = """<noData>
        <noDataList bandNo="1" useSrcNoData="0"/>
        <noDataList bandNo="2" useSrcNoData="0"/>
      </noData>"""

            renderer = """
           <rasterrenderer opacity="1" type="singlebandgray" grayBand="1" nodataColor="" alphaBand="-1" gradient="BlackToWhite">
          <rasterTransparency/>
          <minMaxOrigin>
            <limits>MinMax</limits>
            <extent>WholeRaster</extent>
            <statAccuracy>Estimated</statAccuracy>
            <cumulativeCutLower>0.02</cumulativeCutLower>
            <cumulativeCutUpper>0.98</cumulativeCutUpper>
            <stdDevFactor>2</stdDevFactor>
          </minMaxOrigin>
          <contrastEnhancement>
            <minValue>6</minValue>
            <maxValue>255</maxValue>
            <algorithm>StretchToMinimumMaximum</algorithm>
          </contrastEnhancement>
        </rasterrenderer> 
            
            """
        else:
            nodata = """<noData>
        <noDataList bandNo="1" useSrcNoData="0"/>
        <noDataList bandNo="2" useSrcNoData="0"/>
        <noDataList bandNo="3" useSrcNoData="0"/>
        <noDataList bandNo="4" useSrcNoData="0"/>
      </noData>"""

            renderer = """
            <rasterrenderer type="multibandcolor" alphaBand="4" opacity="1" redBand="1" blueBand="3" nodataColor="" greenBand="2">
          <rasterTransparency/>
          <minMaxOrigin>
            <limits>MinMax</limits>
            <extent>WholeRaster</extent>
            <statAccuracy>Estimated</statAccuracy>
            <cumulativeCutLower>0.02</cumulativeCutLower>
            <cumulativeCutUpper>0.98</cumulativeCutUpper>
            <stdDevFactor>2</stdDevFactor>
          </minMaxOrigin>
          <redContrastEnhancement>
            <minValue>0</minValue>
            <maxValue>255</maxValue>
            <algorithm>NoEnhancement</algorithm>
          </redContrastEnhancement>
          <greenContrastEnhancement>
            <minValue>0</minValue>
            <maxValue>255</maxValue>
            <algorithm>NoEnhancement</algorithm>
          </greenContrastEnhancement>
          <blueContrastEnhancement>
            <minValue>0</minValue>
            <maxValue>255</maxValue>
            <algorithm>NoEnhancement</algorithm>
          </blueContrastEnhancement>
        </rasterrenderer> 
            """

        return f"""
        <maplayer labelsEnabled="1" maxScale="0" autoRefreshTime="0" type="raster" hasScaleBasedVisibilityFlag="0" autoRefreshEnabled="0" styleCategories="AllStyleCategories" minScale="1e+08" refreshOnNotifyEnabled="0" refreshOnNotifyMessage="">
      <extent>
        <xmin>{self.tile.cell.bounds[0] - 0.30517578125 * 256} </xmin>
        <ymin>{self.tile.cell.bounds[1] - 0.30517578125 * 256}</ymin>
        <xmax>{self.tile.cell.bounds[2] + 0.30517578125 * 256}</xmax>
        <ymax>{self.tile.cell.bounds[3] + 0.30517578125 * 256 }</ymax>
      </extent>
      <id>{self.tile.acq_id}_visual_tif_{self.uuid}</id>
      <datasource>{vsify(self.tile, 'visual', self.public)}</datasource>
      <keywordList>
        <value></value>
      </keywordList>
      <layername>{self.name}</layername>
      <srs>
        <spatialrefsys>
          <wkt>{self.proj.crs.to_wkt()}</wkt>
          <proj4>{self.proj.srs}</proj4>
          <srsid>3094</srsid>
          <srid>{self.proj.crs.to_epsg()}</srid>
          <authid>{self.proj.crs.to_string()}</authid>
          <description>{self.proj.crs.name}</description>
          <projectionacronym>utm</projectionacronym>
          <ellipsoidacronym>WGS84</ellipsoidacronym>
          <geographicflag>false</geographicflag>
        </spatialrefsys>
      </srs>
      <resourceMetadata>
        <identifier></identifier>
        <parentidentifier></parentidentifier>
        <language></language>
        <type></type>
        <title></title>
        <abstract></abstract>
        <links/>
        <fees></fees>
        <encoding></encoding>
        <crs>
          <spatialrefsys>
            <wkt>{self.proj.crs.to_wkt()}</wkt>
            <proj4>{self.proj.srs}</proj4>
            <srsid>3094</srsid>
            <srid>{self.proj.crs.to_epsg()}</srid>
            <authid>{self.proj.crs.to_string()}</authid>
            <description>{self.proj.crs.name}</description>
            <projectionacronym>utm</projectionacronym>
            <ellipsoidacronym>WGS84</ellipsoidacronym>
            <geographicflag>false</geographicflag>
            </spatialrefsys>
        </crs>
        <extent/>
      </resourceMetadata>
      <provider>gdal</provider>
     {nodata} 
      <map-layer-style-manager current="default">
        <map-layer-style name="default"/>
      </map-layer-style-manager>
      <flags>
        <Identifiable>1</Identifiable>
        <Removable>1</Removable>
        <Searchable>1</Searchable>
      </flags>
      <customproperties>
        <property key="identify/format" value="Value"/>
      </customproperties>
      <pipe>
        {renderer} 
        <brightnesscontrast brightness="0" contrast="0"/>
        <huesaturation colorizeOn="0" colorizeGreen="128" saturation="0" grayscaleMode="0" colorizeRed="255" colorizeStrength="100" colorizeBlue="128"/>
        <rasterresampler maxOversampling="2"/>
      </pipe>
      <blendMode>0</blendMode>
    </maplayer>"""


##############


class EmptyMask(Exception):
    pass


class MaskDataLayer:
    def __init__(self, tile, asset, color, uuid, public):
        # asset = 'cloud-mask' i.e. asset key
        self.tile = tile
        self.asset = asset
        # color is r,g,b,a, clouds is 229,182,54,255
        self.color = color
        title = asset.replace("-", " ").title().replace("Ms ", "MS ")
        self.name = f"{self.tile.acq_id} {title}"
        self.uuid = uuid
        self.proj = Proj(tile.cell.proj)
        self.public = public
        asset = tile.asset_paths[asset]
        # cloud layer names got changed, have to look it up for now
        with fiona.open(f"{tile.gdal_prefix}{asset}") as layer:
            self.layername = layer.name
            features = list(layer)
            if len(features) == 0:
                raise EmptyMask
            if len(features) == 1:
                if len(features[0]["geometry"]["coordinates"]) == 0:
                    raise EmptyMask

    def as_treelayer(self):

        return f"""<layer-tree-layer checked="Qt::Checked" 
            id="{self.layername}_{self.uuid}" 
            expanded="1" name="{self.name}" 
            legend_exp="" providerKey="ogr" 
            source="{vsify(self.tile, self.asset, self.public)}|layername={self.layername}">
        <customproperties/>
        </layer-tree-layer>"""

    def as_maplayer(self):
        return f"""<maplayer refreshOnNotifyEnabled="0" simplifyAlgorithm="0" simplifyMaxScale="1" autoRefreshTime="0" minScale="100000000" wkbType="MultiPolygon" simplifyDrawingHints="1" readOnly="0" maxScale="0" refreshOnNotifyMessage="" styleCategories="AllStyleCategories" simplifyDrawingTol="1" labelsEnabled="0" type="vector" simplifyLocal="1" hasScaleBasedVisibilityFlag="0" autoRefreshEnabled="0" geometry="Polygon">
      <id>{self.layername}_{self.uuid}</id>
      <datasource>{vsify(self.tile, self.asset, self.public)}|layername={self.layername}</datasource>
      <layername>{self.name}</layername>
      <srs>
        <spatialrefsys>
          <wkt>{self.proj.crs.to_wkt()}</wkt>
          <proj4>{self.proj.srs}</proj4>
          <srsid>3094</srsid>
          <srid>{self.proj.crs.to_epsg()}</srid>
          <authid>{self.proj.crs.to_string()}</authid>
          <description>{self.proj.crs.name}</description>
          <projectionacronym>utm</projectionacronym>
          <ellipsoidacronym>WGS84</ellipsoidacronym>
          <geographicflag>false</geographicflag> 
        </spatialrefsys>
      </srs>
      <provider encoding="UTF-8">ogr</provider>
      <legend type="default-vector"/>
      <map-layer-style-manager current="default">
        <map-layer-style name="default"/>
      </map-layer-style-manager>
      <flags>
        <Identifiable>1</Identifiable>
        <Removable>1</Removable>
        <Searchable>1</Searchable>
      </flags>
      <renderer-v2 type="singleSymbol" enableorderby="0" forceraster="0" symbollevels="0">
        <symbols>
          <symbol type="fill" alpha="1" clip_to_extent="1" force_rhr="0" name="0">
            <layer locked="0" enabled="1" pass="0" class="SimpleFill">
              <prop k="border_width_map_unit_scale" v="3x:0,0,0,0,0,0"/>
              <prop k="color" v="{self.color}"/>
              <prop k="joinstyle" v="bevel"/>
              <prop k="offset" v="0,0"/>
              <prop k="offset_map_unit_scale" v="3x:0,0,0,0,0,0"/>
              <prop k="offset_unit" v="MM"/>
              <prop k="outline_color" v="35,35,35,255"/>
              <prop k="outline_style" v="solid"/>
              <prop k="outline_width" v="0.26"/>
              <prop k="outline_width_unit" v="MM"/>
              <prop k="style" v="solid"/>
              <data_defined_properties>
                <Option type="Map">
                  <Option type="QString" value="" name="name"/>
                  <Option name="properties"/>
                  <Option type="QString" value="collection" name="type"/>
                </Option>
              </data_defined_properties>
            </layer>
          </symbol>
        </symbols>
      </renderer-v2>
      <blendMode>0</blendMode>
      <featureBlendMode>0</featureBlendMode>
      <layerOpacity>1</layerOpacity>
      <geometryOptions geometryPrecision="0" removeDuplicateNodes="0">
        <activeChecks type="StringList">
          <Option type="QString" value=""/>
        </activeChecks>
        <checkConfiguration/>
      </geometryOptions>
      <attributeactions>
        <defaultAction key="Canvas" value="{00000000-0000-0000-0000-000000000000}"/>
      </attributeactions>
    </maplayer> """


class TreeLayer:
    def __init__(self, layers):
        self.layers = layers

    def as_text(self):

        layers = "\n".join(layer.as_treelayer() for layer in self.layers)
        return f"""
            <layer-tree-group checked="Qt::Checked" expanded="1" name="{self.layers[0].tile.cell.id}">
                {layers}
            </layer-tree-group> """


class MapLayer:
    def __init__(self, layers):
        self.layers = layers

    def as_text(self):
        return "\n".join(layer.as_maplayer() for layer in self.layers)


class QLayer:
    def __init__(self, collection, public=False):
        if not HAS_FIONA:
            raise MissingDependency("QLR export requires Fiona")
        self.treelayers = defaultdict(list)
        self.maplayers = []
        self.zone_ids = [(z, uuid4().hex) for z in collection.zones]
        for acq in collection.acquisitions:
            for tile in acq:
                uuid = uuid4().hex
                item = ItemLayer(tile, uuid, public)
                layers = [item]
                try:
                    layers.append(
                        MaskDataLayer(tile, "cloud-mask", "229,182,54,255", uuid, public)
                    )
                except EmptyMask:
                    pass

                # V3 ARD
                if "water-mask" in tile.asset_paths:
                    v3layers = (
                        ("cloud-shadow-mask", "220,67,0,255"),
                        ("water-mask", "0,204,204,128"),
                        ("terrain-shadow-mask", "255,51,51,255"),
                        ("ms-saturation-mask", "255,153,153,255"),
                    )
                    for asset, color in v3layers:
                        try:
                            layers.append(MaskDataLayer(tile, asset, color, uuid, public))
                        except EmptyMask:
                            pass

                visual = VisualDataLayer(tile, uuid, public)
                layers.append(visual)
                self.treelayers[acq].append(TreeLayer(layers))
                self.maplayers.append(MapLayer(layers))

    def render_tree(self):
        text = ""
        for acq, layers in sorted(
            self.treelayers.items(), key=lambda kv: kv[0].properties["datetime"], reverse=True
        ):
            text += f'<layer-tree-group checked="Qt::Unchecked" expanded="0" name="{acq.date}: {acq.acq_id}">'
            text += "<customproperties/>"
            text += "\n".join(t.as_text() for t in layers)
            text += " </layer-tree-group>"
        return text

    def render_layers(self):
        return "\n".join([l.as_text() for l in self.maplayers])

    def render_zones_tree(self):
        text = f'<layer-tree-group checked="Qt::Unchecked" expanded="0" name="ARD Grids">'
        text += "<customproperties/>"
        for zone, uuid in self.zone_ids:
            text += f"""
            <layer-tree-layer legend_exp="" source="/vsicurl/https://ard-docs.s3.amazonaws.com/main/canvas-grids/Z{zone}.gpkg|layername=Z{zone}" providerKey="ogr" expanded="1" name="ARD Grid Zone {zone}" checked="Qt::Unchecked" id="Z{zone}_{uuid}">
            <customproperties/>
            </layer-tree-layer>
            """
        text += " </layer-tree-group>"
        return text

    def render_zones_layers(self):
        text = ""
        for zone, uuid in self.zone_ids:
            proj = Proj(f"EPSG:326{zone:02}")
            text += f"""
            <maplayer simplifyDrawingHints="1" refreshOnNotifyMessage="" simplifyLocal="1" autoRefreshTime="0" simplifyDrawingTol="1" readOnly="0" type="vector" simplifyAlgorithm="0" styleCategories="AllStyleCategories" maxScale="0" minScale="100000000" simplifyMaxScale="1" labelsEnabled="1" wkbType="Polygon" geometry="Polygon" refreshOnNotifyEnabled="0" hasScaleBasedVisibilityFlag="0" autoRefreshEnabled="0">
      <extent>
        <xmin>165000</xmin>
        <ymin>-9330000</ymin>
        <xmax>835000</xmax>
        <ymax>9330000</ymax>
      </extent>
      <id>Z{zone}_{uuid}</id>
      <datasource>/vsicurl/https://ard-docs.s3.amazonaws.com/main/canvas-grids/Z{zone}.gpkg|layername=Z{zone}</datasource>
      <keywordList>
        <value></value>
      </keywordList>
      <layername>ARD Grid Zone {zone}</layername>
      <srs>
       <spatialrefsys>
          <wkt>{proj.crs.to_wkt()}</wkt>
          <proj4>{proj.srs}</proj4>
          <srsid>3094</srsid>
          <srid>{proj.crs.to_epsg()}</srid>
          <authid>{proj.crs.to_string()}</authid>
          <description>{proj.crs.name}</description>
          <projectionacronym>utm</projectionacronym>
          <ellipsoidacronym>WGS84</ellipsoidacronym>
          <geographicflag>false</geographicflag> 
        </spatialrefsys> 
      </srs>
      <provider encoding="UTF-8">ogr</provider>
      <legend type="default-vector"/>
      <map-layer-style-manager current="default">
        <map-layer-style name="default"/>
      </map-layer-style-manager>
      <flags>
        <Identifiable>1</Identifiable>
        <Removable>1</Removable>
        <Searchable>1</Searchable>
      </flags>
      <renderer-v2 type="singleSymbol" forceraster="0" symbollevels="0" enableorderby="0">
        <symbols>
          <symbol type="fill" force_rhr="0" alpha="1" name="0" clip_to_extent="1">
            <layer pass="0" class="SimpleFill" enabled="1" locked="0">
              <prop k="border_width_map_unit_scale" v="3x:0,0,0,0,0,0"/>
              <prop k="color" v="0,0,255,255"/>
              <prop k="joinstyle" v="bevel"/>
              <prop k="offset" v="0,0"/>
              <prop k="offset_map_unit_scale" v="3x:0,0,0,0,0,0"/>
              <prop k="offset_unit" v="MM"/>
              <prop k="outline_color" v="227,69,72,255"/>
              <prop k="outline_style" v="solid"/>
              <prop k="outline_width" v="0.66"/>
              <prop k="outline_width_unit" v="MM"/>
              <prop k="style" v="no"/>
              <data_defined_properties>
                <Option type="Map">
                  <Option type="QString" value="" name="name"/>
                  <Option name="properties"/>
                  <Option type="QString" value="collection" name="type"/>
                </Option>
              </data_defined_properties>
            </layer>
          </symbol>
        </symbols>
        <rotation/>
        <sizescale/>
      </renderer-v2>
      <labeling type="simple">
        <settings calloutType="simple">
          <text-style fontSizeMapUnitScale="3x:0,0,0,0,0,0" fontLetterSpacing="0" useSubstitutions="0" textOrientation="horizontal" fontCapitals="0" fontFamily=".SF NS Text" fontKerning="1" blendMode="0" fontStrikeout="0" previewBkgrdColor="255,255,255,255" fontSize="10" fontItalic="0" multilineHeight="1" fieldName="utmzone + '-' + quadkey" fontWeight="50" fontUnderline="0" textOpacity="1" namedStyle="Regular" isExpression="1" fontSizeUnit="Point" textColor="0,0,0,255" fontWordSpacing="0">
            <text-buffer bufferDraw="1" bufferNoFill="0" bufferJoinStyle="128" bufferSizeMapUnitScale="3x:0,0,0,0,0,0" bufferBlendMode="0" bufferOpacity="1" bufferColor="255,255,255,255" bufferSize="1" bufferSizeUnits="MM"/>
            <text-mask maskOpacity="1" maskEnabled="0" maskJoinStyle="128" maskSizeMapUnitScale="3x:0,0,0,0,0,0" maskSize="1.5" maskType="0" maskSizeUnits="MM" maskedSymbolLayers=""/>
            <background shapeDraw="0" shapeSizeX="0" shapeOpacity="1" shapeOffsetUnit="MM" shapeOffsetMapUnitScale="3x:0,0,0,0,0,0" shapeOffsetY="0" shapeOffsetX="0" shapeSizeMapUnitScale="3x:0,0,0,0,0,0" shapeRadiiX="0" shapeRadiiUnit="MM" shapeFillColor="255,255,255,255" shapeBorderWidth="0" shapeJoinStyle="64" shapeBlendMode="0" shapeBorderWidthMapUnitScale="3x:0,0,0,0,0,0" shapeSizeType="0" shapeRadiiMapUnitScale="3x:0,0,0,0,0,0" shapeBorderColor="128,128,128,255" shapeSVGFile="" shapeSizeUnit="MM" shapeRotation="0" shapeType="0" shapeSizeY="0" shapeRotationType="0" shapeBorderWidthUnit="MM" shapeRadiiY="0">
              <symbol type="marker" force_rhr="0" alpha="1" name="markerSymbol" clip_to_extent="1">
                <layer pass="0" class="SimpleMarker" enabled="1" locked="0">
                  <prop k="angle" v="0"/>
                  <prop k="color" v="196,60,57,255"/>
                  <prop k="horizontal_anchor_point" v="1"/>
                  <prop k="joinstyle" v="bevel"/>
                  <prop k="name" v="circle"/>
                  <prop k="offset" v="0,0"/>
                  <prop k="offset_map_unit_scale" v="3x:0,0,0,0,0,0"/>
                  <prop k="offset_unit" v="MM"/>
                  <prop k="outline_color" v="35,35,35,255"/>
                  <prop k="outline_style" v="solid"/>
                  <prop k="outline_width" v="0"/>
                  <prop k="outline_width_map_unit_scale" v="3x:0,0,0,0,0,0"/>
                  <prop k="outline_width_unit" v="MM"/>
                  <prop k="scale_method" v="diameter"/>
                  <prop k="size" v="2"/>
                  <prop k="size_map_unit_scale" v="3x:0,0,0,0,0,0"/>
                  <prop k="size_unit" v="MM"/>
                  <prop k="vertical_anchor_point" v="1"/>
                  <data_defined_properties>
                    <Option type="Map">
                      <Option type="QString" value="" name="name"/>
                      <Option name="properties"/>
                      <Option type="QString" value="collection" name="type"/>
                    </Option>
                  </data_defined_properties>
                </layer>
              </symbol>
            </background>
            <shadow shadowRadius="1.5" shadowUnder="0" shadowRadiusMapUnitScale="3x:0,0,0,0,0,0" shadowColor="0,0,0,255" shadowDraw="0" shadowOffsetAngle="135" shadowOffsetDist="1" shadowScale="100" shadowBlendMode="6" shadowRadiusAlphaOnly="0" shadowOffsetGlobal="1" shadowRadiusUnit="MM" shadowOpacity="0.7" shadowOffsetUnit="MM" shadowOffsetMapUnitScale="3x:0,0,0,0,0,0"/>
            <dd_properties>
              <Option type="Map">
                <Option type="QString" value="" name="name"/>
                <Option name="properties"/>
                <Option type="QString" value="collection" name="type"/>
              </Option>
            </dd_properties>
            <substitutions/>
          </text-style>
          <text-format multilineAlign="0" addDirectionSymbol="0" reverseDirectionSymbol="0" autoWrapLength="0" formatNumbers="0" plussign="0" wrapChar="" rightDirectionSymbol=">" decimals="3" leftDirectionSymbol="&lt;" placeDirectionSymbol="0" useMaxLineLengthForAutoWrap="1"/>
          <placement fitInPolygonOnly="0" distMapUnitScale="3x:0,0,0,0,0,0" rotationAngle="0" labelOffsetMapUnitScale="3x:0,0,0,0,0,0" maxCurvedCharAngleIn="25" predefinedPositionOrder="TR,TL,BR,BL,R,L,TSR,BSR" distUnits="MM" placement="0" placementFlags="10" geometryGeneratorEnabled="0" preserveRotation="1" geometryGenerator="" quadOffset="0" offsetUnits="MM" priority="5" geometryGeneratorType="PointGeometry" offsetType="0" centroidWhole="1" repeatDistance="0" dist="0" xOffset="0" yOffset="0" centroidInside="0" repeatDistanceUnits="MM" repeatDistanceMapUnitScale="3x:0,0,0,0,0,0" maxCurvedCharAngleOut="-25" layerType="PolygonGeometry" overrunDistance="0" overrunDistanceUnit="MM" overrunDistanceMapUnitScale="3x:0,0,0,0,0,0"/>
          <rendering fontMinPixelSize="3" minFeatureSize="0" drawLabels="1" displayAll="0" scaleVisibility="1" maxNumLabels="2000" mergeLines="0" obstacleFactor="1" obstacleType="1" labelPerPart="0" scaleMin="0" upsidedownLabels="0" obstacle="1" zIndex="0" scaleMax="134345" fontMaxPixelSize="10000" fontLimitPixelSize="0" limitNumLabels="0"/>
          <dd_properties>
            <Option type="Map">
              <Option type="QString" value="" name="name"/>
              <Option name="properties"/>
              <Option type="QString" value="collection" name="type"/>
            </Option>
          </dd_properties>
          <callout type="simple">
            <Option type="Map">
              <Option type="QString" value="pole_of_inaccessibility" name="anchorPoint"/>
              <Option type="Map" name="ddProperties">
                <Option type="QString" value="" name="name"/>
                <Option name="properties"/>
                <Option type="QString" value="collection" name="type"/>
              </Option>
              <Option type="bool" value="false" name="drawToAllParts"/>
              <Option type="QString" value="0" name="enabled"/>
              <Option type="QString" value="&lt;symbol type=&quot;line&quot; force_rhr=&quot;0&quot; alpha=&quot;1&quot; name=&quot;symbol&quot; clip_to_extent=&quot;1&quot;>&lt;layer pass=&quot;0&quot; class=&quot;SimpleLine&quot; enabled=&quot;1&quot; locked=&quot;0&quot;>&lt;prop k=&quot;capstyle&quot; v=&quot;square&quot;/>&lt;prop k=&quot;customdash&quot; v=&quot;5;2&quot;/>&lt;prop k=&quot;customdash_map_unit_scale&quot; v=&quot;3x:0,0,0,0,0,0&quot;/>&lt;prop k=&quot;customdash_unit&quot; v=&quot;MM&quot;/>&lt;prop k=&quot;draw_inside_polygon&quot; v=&quot;0&quot;/>&lt;prop k=&quot;joinstyle&quot; v=&quot;bevel&quot;/>&lt;prop k=&quot;line_color&quot; v=&quot;60,60,60,255&quot;/>&lt;prop k=&quot;line_style&quot; v=&quot;solid&quot;/>&lt;prop k=&quot;line_width&quot; v=&quot;0.3&quot;/>&lt;prop k=&quot;line_width_unit&quot; v=&quot;MM&quot;/>&lt;prop k=&quot;offset&quot; v=&quot;0&quot;/>&lt;prop k=&quot;offset_map_unit_scale&quot; v=&quot;3x:0,0,0,0,0,0&quot;/>&lt;prop k=&quot;offset_unit&quot; v=&quot;MM&quot;/>&lt;prop k=&quot;ring_filter&quot; v=&quot;0&quot;/>&lt;prop k=&quot;use_custom_dash&quot; v=&quot;0&quot;/>&lt;prop k=&quot;width_map_unit_scale&quot; v=&quot;3x:0,0,0,0,0,0&quot;/>&lt;data_defined_properties>&lt;Option type=&quot;Map&quot;>&lt;Option type=&quot;QString&quot; value=&quot;&quot; name=&quot;name&quot;/>&lt;Option name=&quot;properties&quot;/>&lt;Option type=&quot;QString&quot; value=&quot;collection&quot; name=&quot;type&quot;/>&lt;/Option>&lt;/data_defined_properties>&lt;/layer>&lt;/symbol>" name="lineSymbol"/>
              <Option type="double" value="0" name="minLength"/>
              <Option type="QString" value="3x:0,0,0,0,0,0" name="minLengthMapUnitScale"/>
              <Option type="QString" value="MM" name="minLengthUnit"/>
              <Option type="double" value="0" name="offsetFromAnchor"/>
              <Option type="QString" value="3x:0,0,0,0,0,0" name="offsetFromAnchorMapUnitScale"/>
              <Option type="QString" value="MM" name="offsetFromAnchorUnit"/>
              <Option type="double" value="0" name="offsetFromLabel"/>
              <Option type="QString" value="3x:0,0,0,0,0,0" name="offsetFromLabelMapUnitScale"/>
              <Option type="QString" value="MM" name="offsetFromLabelUnit"/>
            </Option>
          </callout>
        </settings>
      </labeling>
      <customproperties>
        <property value="0" key="embeddedWidgets/count"/>
        <property key="variableNames"/>
        <property key="variableValues"/>
      </customproperties>
      <blendMode>0</blendMode>
      <featureBlendMode>0</featureBlendMode>
      <layerOpacity>1</layerOpacity>
      <SingleCategoryDiagramRenderer diagramType="Histogram" attributeLegend="1">
        <DiagramCategory sizeType="MM" backgroundAlpha="255" penColor="#000000" sizeScale="3x:0,0,0,0,0,0" penWidth="0" spacingUnit="MM" showAxis="1" width="15" rotationOffset="270" spacingUnitScale="3x:0,0,0,0,0,0" enabled="0" opacity="1" diagramOrientation="Up" barWidth="5" maxScaleDenominator="1e+08" labelPlacementMethod="XHeight" lineSizeType="MM" minimumSize="0" spacing="5" scaleBasedVisibility="0" direction="0" minScaleDenominator="0" penAlpha="255" lineSizeScale="3x:0,0,0,0,0,0" scaleDependency="Area" backgroundColor="#ffffff" height="15">
          <fontProperties description=".SF NS Text,13,-1,5,50,0,0,0,0,0" style=""/>
          <axisSymbol>
            <symbol type="line" force_rhr="0" alpha="1" name="" clip_to_extent="1">
              <layer pass="0" class="SimpleLine" enabled="1" locked="0">
                <prop k="capstyle" v="square"/>
                <prop k="customdash" v="5;2"/>
                <prop k="customdash_map_unit_scale" v="3x:0,0,0,0,0,0"/>
                <prop k="customdash_unit" v="MM"/>
                <prop k="draw_inside_polygon" v="0"/>
                <prop k="joinstyle" v="bevel"/>
                <prop k="line_color" v="35,35,35,255"/>
                <prop k="line_style" v="solid"/>
                <prop k="line_width" v="0.26"/>
                <prop k="line_width_unit" v="MM"/>
                <prop k="offset" v="0"/>
                <prop k="offset_map_unit_scale" v="3x:0,0,0,0,0,0"/>
                <prop k="offset_unit" v="MM"/>
                <prop k="ring_filter" v="0"/>
                <prop k="use_custom_dash" v="0"/>
                <prop k="width_map_unit_scale" v="3x:0,0,0,0,0,0"/>
                <data_defined_properties>
                  <Option type="Map">
                    <Option type="QString" value="" name="name"/>
                    <Option name="properties"/>
                    <Option type="QString" value="collection" name="type"/>
                  </Option>
                </data_defined_properties>
              </layer>
            </symbol>
          </axisSymbol>
        </DiagramCategory>
      </SingleCategoryDiagramRenderer>
      <DiagramLayerSettings linePlacementFlags="18" zIndex="0" dist="0" placement="1" priority="0" showAll="1" obstacle="0">
        <properties>
          <Option type="Map">
            <Option type="QString" value="" name="name"/>
            <Option name="properties"/>
            <Option type="QString" value="collection" name="type"/>
          </Option>
        </properties>
      </DiagramLayerSettings>
      <geometryOptions geometryPrecision="0" removeDuplicateNodes="0">
        <activeChecks/>
        <checkConfiguration type="Map">
          <Option type="Map" name="QgsGeometryGapCheck">
            <Option type="double" value="0" name="allowedGapsBuffer"/>
            <Option type="bool" value="false" name="allowedGapsEnabled"/>
            <Option type="QString" value="" name="allowedGapsLayer"/>
          </Option>
        </checkConfiguration>
      </geometryOptions>
     
      <attributeactions>
        <defaultAction value="{00000000-0000-0000-0000-000000000000}" key="Canvas"/>
      </attributeactions>
    </maplayer>"""

        return text

    def as_text(self):
        return f"""<!DOCTYPE qgis-layer-definition>
        <qlr>
           <layer-tree-group checked="Qt::Checked" expanded="1" name="">
            <customproperties/>
            {self.render_zones_tree()}
            {self.render_tree()}
            </layer-tree-group> 
            <maplayers>
            {self.render_zones_layers()}
            {self.render_layers()}
            </maplayers>
        </qlr> """



""" Authenticated sessions for communicating with the ARD API endpoints"""

import base64
import json
import os
import warnings

from configparser import ConfigParser
from oauthlib.oauth2 import BackendApplicationClient, LegacyApplicationClient
from oauthlib.oauth2.rfc6749.errors import MissingTokenError
from platform import python_version, system
from requests.auth import HTTPBasicAuth
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from requests_oauthlib import OAuth2Session

from max_ard.exceptions import (
    ard_server_request_hook,
    bad_ard_request_hook,
    bad_geom_request_hook,
    oversize_request_hook,
    unauth_request_hook,
    missing_resource_hook,
)


try:
    from importlib import metadata
except ImportError:  # for Python<3.8
    import importlib_metadata as metadata

__all__ = ("get_user_session", "get_client_session", "get_self")

SAVE_TOKEN = True  # if false, never save the token back to the config file
ARD_TIMEOUT = os.environ.get(
    "ARD_TIMEOUT", 30
)  # seconds, this is high for big selects and dry run orders


def jwt_expires(token):
    """returns the expiration from a JWT token

    Arguments
    ---------
    token: str
        JWT token to decode

    Returns
    -------
    str: expiration time

    Notes
    -----
    THIS DOES NOT VERIFY THE TOKEN"""

    payload_segment = token.split(".")[1]
    input = payload_segment.encode("ascii")
    rem = len(input) % 4
    if rem > 0:
        input += b"=" * (4 - rem)
    payload = base64.urlsafe_b64decode(input)
    return json.loads(payload)["exp"]


# Timeout and Retry handling
# from https://findwork.dev/blog/advanced-usage-python-requests-timeouts-retries-hooks/


class TimeoutHTTPAdapter(HTTPAdapter):
    """HTTP Adapter that adds in default timeouts

    Keywords
    --------
    timeout(numeric): seconds to wait before timing out

    """

    def __init__(self, *args, **kwargs):
        self.timeout = ARD_TIMEOUT
        if "timeout" in kwargs:
            self.timeout = kwargs["timeout"]
            del kwargs["timeout"]
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        timeout = kwargs.get("timeout")
        if timeout is None:
            kwargs["timeout"] = self.timeout
        return super().send(request, **kwargs)


def mount_adaptors(session):
    """Mount adaptors to a session for retries and update UA string

    Arguments
    ---------
    session: Requests session
    """

    retries = Retry(total=3, backoff_factor=5, status_forcelist=[429, 503, 504])
    session.mount("https://", TimeoutHTTPAdapter(max_retries=retries))
    session.mount("http://", TimeoutHTTPAdapter(max_retries=retries))
    version = metadata.version("max_ard")
    session.headers.update(
        {"User-Agent": f"max_ard/{version} (Python {python_version()}/{system()})"}
    )
    return session


def link_paginated_response(session, url, **params):
    """Follow responses with paginated links and gather up objects

    Newer paginated endpoints give you a link to retrieve the next set of objects.
    For older endpoints that only return object IDs, see the next function.

    Parameters
    ----------
    session: Requests session
        A session through which to make the requests, usually an ARD session
    url: str
        URL to start fetching objects
    **params: any
        Additional parameters to pass through to the session's GET method

    Returns
    -------
    list: any
        A list of objects returned by the endpoint"""

    response = {"has_more": True}
    things = []
    while response["has_more"]:
        response = session.get(url, params=params).json()
        for thing in response["data"]:
            things.append(thing)
        # nothing returned
        if len(things) == 0:
            return []
        if response["has_more"]:
            url = response["links"]["next_page"]

    return things


def paginated_response(session, url, key, **params):
    """Follow paginated responses with object IDs and gather up objects

    This uses `starting_after` to grab the next batch of objects after the last
    object's ID.

    Parameters
    ----------
    session: Requests session
        A session through which to make the requests, usually an ARD session
    url: str
        URL to start fetching objects
    key: callable
        A callable that given the object, returns the object's ID.
    **params: any
        Additional parameters to pass through to the session's GET method

    Returns
    -------
    list: any
        A list of objects returned by the endpoint"""

    response = {"has_more": True}
    things = []
    while response["has_more"]:
        response = session.get(url, params=params).json()
        for thing in response["data"]:
            things.append(thing)
            last_id = key(thing)
        # nothing returned
        if len(things) == 0:
            return []
        params["starting_after"] = last_id

    return things


def ard_url(*args):
    """Get an ARD url built from given subfolders

    This will build the url using hostname and dev feature branch envvars, if present.

    Parameters
    ----------
    *args: any
        ARD API endpoint subfolders to concatenate into a url

    Returns
    -------
    str: endpoint url

    Examples
    --------

    No hostname or branch set returns prod urls:

        ard_url('select', 'status') -> https://ard.maxar.com/api/v1/select/status

    Hostname set:

        MAXAR_ARD_HOSTNAME = ard-dev.maxar.com

        ard_url('select', 'status') -> https://ard-dev.maxar.com/api/v1/select/status

    Feature Branch (only applies to dev):

        MAXAR_ARD_HOSTNAME = ard-dev.maxar.com
        ARD_DEV_BRANCH = new_kml

        ard_url('select', 'status') -> https://ard-dev.maxar.com/api/v1/select-new_kml/status

    """
    hostname = os.environ.get("MAXAR_ARD_HOSTNAME", "ard.maxar.com/api/v1")
    dev_branch = os.environ.get("ARD_DEV_BRANCH")
    if dev_branch:
        return f'https://{hostname}/{args[0]}-{dev_branch}/{"/".join(args[1:])}'
    else:
        return f'https://{hostname}/{args[0]}/{"/".join(args[1:])}'


AUTH_URL = ard_url("auth", "authenticate")

_USER_SESSION_CACHE = None
_CLIENT_SESSION_CACHE = {}


def get_self(session):
    """Fetch the account SELF endpoint of the current user

    Parameters
    ----------
    session : OAuth2Session
        The session object created for the current user

    Returns
    -------
    dict: SELF api payload"""

    r = session.get(ard_url("auth", "self"))
    r.raise_for_status()
    return r.json()


def get_session():
    warnings.warn(
        "get_session() will be deprecated for the more specific get_user_session()",
        DeprecationWarning,
    )
    return get_user_session()


def add_session_hooks(session):
    """
    Add hooks for session object, and return it.
    Note: Order of Hooks is important.
    """
    session.hooks["response"] = [
        # Specific 400 errors
        bad_geom_request_hook,
        # Generic 400 errors
        bad_ard_request_hook,
        # 413 errors
        oversize_request_hook,
        # 403 errors
        unauth_request_hook,
        # 404 errors
        missing_resource_hook,
        # non-200 errors
        ard_server_request_hook,
    ]

    return session


def get_user_session():
    """Get an authenticated Requests session for communicating with the ARD API

    Example
    -------
    >>> r = get_user_session()
    >>> resp = r.get('https://ard-dev.geobigdata.io/api/v1/select/request/5542092796671610498')

    Returns
    -------
    OAuth2Session
        Authenticated Requests session

    Notes
    -----
    For credentials, create a file in your home directory called .ard-config with these contents:

        [ard]
        user_name = <your_user_name>
        user_password = <your_password>

    or set these environment variables:

        ARD_USERNAME
        ARD_PASSWORD
    """
    global _USER_SESSION_CACHE
    if _USER_SESSION_CACHE is None:
        _USER_SESSION_CACHE = _get_user_session()
    return _USER_SESSION_CACHE


def _get_user_session(config_file=None):
    """Get an authenticated Requests session for communicating with the ARD API

    Parameters
    ----------
    config_file, optional
        Path to config file

    Returns
    -------
    OAuth2Session
        Authenticated Requests session

    Notes
    -----
    If you provide ARD_ACCESS_TOKEN and ARD_REFRESH_TOKEN via env vars it will
    use those credentials.  If you provide a path to a config
    file, it will look there for the credentials. If you don't it will try to
    pull the credentials from environment variables (ARD_USERNAME, ARD_PASSWORD).
    If that fails and you have a '~/.ard-config' ini file, it will read from that.
    """

    session = None
    if os.environ.get("ARD_ACCESS_TOKEN", None):
        session = session_from_existing_token(
            access_token=os.environ.get("ARD_ACCESS_TOKEN", None),
            refresh_token=os.environ.get("ARD_REFRESH_TOKEN", None),
        )

    # If no config file specified, try using environment variables.  If that
    # fails and there is a config in the default location, use that.
    if not session and not config_file:
        try:
            session = session_from_envvars()
        except MissingTokenError as e:
            raise Exception("Invalid ARD credentials given in environment variables.")
        except Exception as e:
            config_file = os.path.expanduser("~/.ard-config")

    example_credentials = """
    [ard]
    user_name = your_user_name
    user_password = your_password
    """

    if config_file and not os.path.isfile(config_file):
        raise Exception(
            f"Please create a ARD credential file at ~/.ard-config with these contents:\n{example_credentials}"
        )

    if not session:
        try:
            session = session_from_config(config_file)
        except:
            raise Exception(
                "Invalid credentials or incorrectly formatted config file at ~/.ard-config"
            )

    session = mount_adaptors(session)
    return add_session_hooks(session)


def get_client_session(client_id=None, client_secret=None):
    """Get an authenticated session using client credentials (Maxar internal only)

    If a client ID and secret are not passed as kwargs, the env vars
    ARD_CLIENT_ID and ARD_CLIENT_SECRET will be used if they exist."""

    global _CLIENT_SESSION_CACHE
    if _CLIENT_SESSION_CACHE[client_id] is None:
        _CLIENT_SESSION_CACHE[client_id] = _get_client_session(
            client_id=client_id, client_secret=client_secret
        )

    return _CLIENT_SESSION_CACHE


def _get_client_session(client_id=None, client_secret=None):
    """Get an authenticated session using client credentials (Maxar internal only)

    If a client ID and secret are not passed as kwargs, the env vars
    ARD_CLIENT_ID and ARD_CLIENT_SECRET will be used if they exist."""

    token_url = f"{AUTH_URL}/oauth2/token"

    if not client_id:
        client_id = os.environ.get("ARD_CLIENT_ID")
    if not client_secret:
        client_secret = os.environ.get("ARD_CLIENT_SECRET")

    if not client_id or not client_secret:
        raise ValueError("Unable to find credentials")

    auth = HTTPBasicAuth(client_id, client_secret)
    client = BackendApplicationClient(client_id=client_id)

    # set up our OAuth session
    oauth = OAuth2Session(
        client=client, scope=["com.maxar.ard/ard.search", "com.maxar.ard/ard.order"]
    )

    token = oauth.fetch_token(token_url=token_url, auth=auth)

    return mount_adaptors(oauth)


def session_from_existing_token(access_token, refresh_token="no_refresh_token", auth_url=AUTH_URL):
    """Get a session using an existing access_token and refresh_token.

    Parameters
    ----------
    access_token : str
        Access token
    refresh_token : str
        Refresh token

    Returns
    -------
    OAuth2Session
        Authenticated Requests session
    """

    def save_token(token):
        s.token = token

    token = {
        "token_type": "Bearer",
        "refresh_token": refresh_token,
        "access_token": access_token,
        "scope": ["read", "write"],
        "expires_in": 604800,
        "expires_at": jwt_expires(access_token),
    }

    s = OAuth2Session(
        token=token,
        auto_refresh_url=auth_url,
        token_updater=save_token,
    )

    return s


def session_from_envvars(
    auth_url=AUTH_URL,
    environ_template=(("username", "ARD_USERNAME"), ("password", "ARD_PASSWORD")),
):
    """Returns a session with the ARD authorization token baked in,
    pulling the credentials from environment variables.

    environ_template - An iterable of key, value pairs. The key should
                      be the variables used in the oauth workflow, and
                      the values being the environment variables to
                      pull the configuration from.  Change the
                      template values if your envvars differ from the
                      default, but make sure the keys remain the same.
    Returns
    -------
    OAuth2Session
        Authenticated Requests session
    """

    def save_token(token):
        s.token = token

    client_id = "dummyclientid"
    client_secret = "dummyclientsecret"

    environ = {var: os.environ[envvar] for var, envvar in environ_template}
    s = OAuth2Session(
        client=LegacyApplicationClient(client_id),
        auto_refresh_url=auth_url,
        auto_refresh_kwargs={"client_id": client_id, "client_secret": client_secret},
        token_updater=save_token,
    )

    s.fetch_token(auth_url, **environ)

    return s


def session_from_kwargs(**kwargs):
    """Get a  session object using credentials from keywords

    Parameters
    ----------
    **kwargs : dict
        Authentication keywords: username, password, client_id, client_secret

    Returns
    -------
    OAuth2Session
        Authenticated Requests session
    """

    def save_token(token):
        s.token = token

    auth_url = AUTH_URL
    s = OAuth2Session(
        client=LegacyApplicationClient(kwargs.get("client_id")),
        auto_refresh_url=auth_url,
        auto_refresh_kwargs={
            "client_id": kwargs.get("client_id"),
            "client_secret": kwargs.get("client_secret"),
        },
        token_updater=save_token,
    )

    try:
        s.fetch_token(
            auth_url,
            username=kwargs.get("username"),
            password=kwargs.get("password"),
            client_id=kwargs.get("client_id"),
            client_secret=kwargs.get("client_secret"),
        )
    except MissingTokenError as e:
        raise Exception("Invalid credentials passed into session_from_kwargs()")

    return s


def session_from_config(config_file):
    """Get a  session object using credentials from a config file

    Parameters
    ----------
    config_file : str
        path to config file

    Returns
    -------
    OAuth2Session
        Authenticated Requests session
    """

    def save_token(token_to_save):
        """Save off the token back to the config file."""
        if not SAVE_TOKEN:
            return
        if not "ard_token" in set(cfg.sections()):
            cfg.add_section("ard_token")
        cfg.set("ard_token", "json", json.dumps(token_to_save))
        with open(config_file, "w") as sink:
            cfg.write(sink)

    # Read the config file (ini format).
    cfg = ConfigParser(interpolation=None)
    if not cfg.read(config_file):
        raise RuntimeError("No ini file found at {} to parse.".format(config_file))
    client_id = "dummy_client_id(not-required)"
    client_secret = "dummy_client_secret(not-required)"

    # the ini file has the optional ability to set an auth url (useful for dev)
    if not cfg.has_option("ard", "auth_url"):
        auth_url = AUTH_URL
    else:
        auth_url = cfg.get("ard", "auth_url")

    # See if we have a token stored in the config, and if not, get one.
    if "ard_token" in set(cfg.sections()):
        # Parse the token from the config.
        token = json.loads(cfg.get("ard_token", "json"))

        s = OAuth2Session(
            client_id,
            client=LegacyApplicationClient(client_id, token=token),
            auto_refresh_url=auth_url,
            auto_refresh_kwargs={"client_id": client_id, "client_secret": client_secret},
            token_updater=save_token,
        )
        s.token = token
        # hit the server to trigger a refresh if the token has expired
        try:
            s.get(ard_url("auth", "self"))
            return s
        except MissingTokenError:
            pass

    # No pre-existing token or the refresh expired, so we request one from the API.
    s = OAuth2Session(
        client_id,
        client=LegacyApplicationClient(client_id),
        auto_refresh_url=auth_url,
        auto_refresh_kwargs={"client_id": client_id, "client_secret": client_secret},
        token_updater=save_token,
    )

    # Get the token and save it to the config.
    token = s.fetch_token(
        auth_url,
        username=cfg.get("ard", "user_name"),
        password=cfg.get("ard", "user_password"),
        client_id=client_id,
        client_secret=client_secret,
        auth=False,
    )
    save_token(token)

    return s


def write_config(username, password):
    """Write a config file

    The config file will be written to ~/.ard-config.

    Parameters
    ----------
    username : str
        Username
    password: str
        Password
    """

    config_file = os.path.expanduser("~/.ard-config")
    cfg = ConfigParser()
    cfg.read(config_file)

    # create or update ard section
    if "ard" not in cfg.sections():
        cfg["ard"] = {}
    if username:
        cfg["ard"]["user_name"] = username
    if password:
        cfg["ard"]["user_password"] = password

    # reset the token
    cfg.remove_section("ard_token")

    with open(config_file, "w") as configfile:
        cfg.write(configfile)


def read_token():
    """Read the token from the config file

    Returns
    -------
    str
        Access token
    """

    config_file = os.path.expanduser("~/.ard-config")
    cfg = ConfigParser()
    cfg.read(config_file)
    print(json.loads(cfg["ard_token"]["json"])["access_token"])



import json
import os
import pprint
import sys
import time
import webbrowser
from collections import defaultdict
from os.path import splitext
from tempfile import NamedTemporaryFile

import click

from max_ard.exceptions import BadARDRequest
from max_ard.io import KmlDoc, KmzDoc, ShpDoc
from max_ard.select import Select


@click.group()
def select():
    """Tools for interacting with the ARD Select service"""
    pass


@click.command()
@click.argument("select_id", required=True)
@click.option(
    "--format",
    help="Return the status in an alternate format. `raw` will return the full response JSON.",
)
def status(select_id, format):
    """Return the status of a Select request using its SELECT_ID"""
    try:
        s = Select.from_id(select_id)
    except Exception as e:
        click.secho(str(e), fg="red", err=True)
        sys.exit()
    if format == "raw":
        click.secho(str(s.response.dict()), fg="cyan")
    else:
        fg_color = "cyan"
        err = False
        if s.state == "SUCCEEDED":
            fg_color = "green"
        elif s.state == "FAILED":
            fg_color = "red"
            err = True
        click.secho(s.state, fg=fg_color, err=err)


@click.command()
@click.argument("select_id", required=True)
@click.option(
    "--dest",
    help="Returns the select in one of the generated format types: `html`, `geojson`, `geojsonl`, `kml`, `kmz`, `shp`, `stac`.",
)
@click.option(
    "--format",
    help="If --dest is not provided, outputs to stdout in these formats:"
    + "`html`, `geojsonl`, `kml`, `stac`. Ignored if --dest is used",
)
@click.option("--verbose", "-v", is_flag=True, help="Return more information about the Select")
def describe(select_id, dest=None, format=None, verbose=False):
    """Returns info about the Select given the SELECT_ID. If no DEST is supplied, prints out a summary of the select.

    Providing a DEST writes to that destination the contents of the given Select result file generated by the system.
    Make sure to include the correct file extension in your desination filepath.
    Formats `geojson`, and `geojsonl`, `kml`, `kmz`, `shp`, and `stac` return files describing the tiles. Format `html` is an interactive map
    browser of the results - to view the map without downloading it first, use `maxard select browse SELECT_ID`"""
    try:
        s = Select.from_id(select_id)
    except Exception as e:
        click.secho(str(e), fg="red", err=True)
        sys.exit()
    pp = pprint.PrettyPrinter(indent=4)
    if not s.finished:
        click.secho("Select has not finished running", fg="cyan")
        if verbose:
            click.secho("*******************", fg="cyan")
            click.secho("Request details:", fg="cyan")
            click.secho(pp.pformat(s.request.to_payload()), fg="cyan")
            click.secho("*******************", fg="cyan")
        exit()
    if s.response.error_message is not None:
        error = s.response.error_message
        click.secho(f'Error in selection process: {error["Error"]}', fg="red", err=True)
        click.secho(f'Reason: {error["Cause"]}', fg="red", err=True)
        exit()
    if dest is None and format is None:
        click.secho("\n")
        click.secho(f"Getting Select ID {select_id}...", fg="cyan")
        click.secho("", fg="cyan")
        click.secho("*******************", fg="cyan")
        click.secho("Request details:", fg="cyan")
        click.secho(pp.pformat(s.request.to_payload()), fg="cyan")
        click.secho("*******************", fg="cyan")
        click.secho("")
        geojson = json.loads(s.get_link_contents("geojson"))
        acqs = defaultdict(list)
        count = 0
        for tile in geojson["features"]:
            for feature in tile["properties"]["best_matches"]:
                count += 1
                acqs[feature["acquisition_id"]].append(feature["date"])
        click.secho(
            f"{count} tiles identified in the following {len(acqs.keys())} acquisitions:",
            fg="green",
        )
        for acq, dates in acqs.items():
            click.secho(f"{acq} ({dates[0]}) - {len(dates)} tiles", fg="green")
        click.secho("")
        click.secho("Ordering this selection will use:", fg="cyan")
        click.secho(f"- {s.usage.area.fresh_imagery_sqkm} sqkm of fresh imagery", fg="cyan")
        click.secho(f"- {s.usage.area.standard_imagery_sqkm} sqkm of standard imagery", fg="cyan")
        click.secho(f"- {s.usage.area.training_imagery_sqkm} sqkm of training imagery", fg="cyan")
        # click.secho('', fg='cyan')
        # click.secho(f'As of {s.usage["usage_as_of"]}:', fg='cyan')
        # click.secho(f'{s.usage["available_sqkm"]} sqkm available', fg='cyan')
        click.secho("")
        exit()
    else:
        if dest:
            format = splitext(dest)[1][1:]

        if format in ["html", "geojson", "geojsonl", "stac", "kml", "kmz", "shp"]:

            # path = Path(dest)

            if (format in s.response.links.keys()) or (
                format == "kml"
            ):  # html, geojson, geojsonl, stac, kml
                if format in s.response.links.keys():
                    file = s.get_link_contents(format)
                else:  # kml
                    file = KmlDoc(s)

                if dest:
                    with open(dest, "w") as out:
                        out.write(file)
                else:
                    click.echo(file)

            if format in ["kmz", "shp"]:
                if dest is None:
                    click.secho(
                        f"Format {format} can not be sent to stdout, use --dest <filename> instead",
                        fg="red",
                        err=True,
                    )
                if format == "kmz":
                    KmzDoc(s, dest)
                else:  # shp
                    ShpDoc(s.get_link_contents("geojson"), dest)

            click.secho(f"Select result file written to {dest}", fg="green")

        else:
            click.secho(
                f'Unknown format {format}, try "html", "geojson", "geojsonl", "kml", "kmz", "shp", or "stac"',
                fg="red",
                err=True,
            )


@click.command()
@click.argument("select_id", required=True)
@click.option(
    "--format",
    help="Format of result file to get signed url for: `html`, `geojson`, geojsonl`, or `stac`",
)
def url(select_id, format):
    """Gets links for the Select result files for a Select of SELECT_ID. If no FORMAT is supplied, returns the base URL of the Select.
    If a FORMAT is supplied, returns a pre-signed URL to download the result of type FORMAT."""
    try:
        s = Select.from_id(select_id)
    except Exception as e:
        click.secho(str(e), fg="red", err=True)
        sys.exit()
    if not s.finished:
        click.secho("Select has not finished running", fg="cyan")
        sys.exit()
    if format is None:
        click.secho(s.response.links["self"], fg="green")
    elif format in s.response.links:
        click.secho(s.get_signed_link(format), fg="green")
    else:
        click.secho(
            f'Unknown format {format}, try "html", "geojson", "geojsonl", or "stac"',
            fg="red",
            err=True,
        )


@click.command()
@click.argument("select_id", required=True)
def browse(select_id):
    """Downloads the HTML interactive viewer of the Select result SELECT ID to a temporary file and opens the
    map with the system web browser.
    When exited the temporary file is deleted. Use `max-ard select describe SELECT_ID --format html > my_map.html`
    to save a local copy"""

    click.secho("Fetching map...", fg="cyan")
    try:
        s = Select.from_id(select_id)
    except Exception as e:
        click.secho(str(e), fg="red", err=True)
        sys.exit()
    if not s.finished:
        click.secho("Select has not finished running", fg="cyan")
        exit()
    prefix = f"ARD_Select_{select_id}-"
    with NamedTemporaryFile(prefix=prefix, suffix=".html", delete=False) as tmp:
        tmp.write(s.get_link_contents("html").encode(encoding="UTF-8"))
    # Windows won't let you open the file from browser while the context manager has it open
    # so we have to manually delete it
    webbrowser.open(f"file://{tmp.name}")
    time.sleep(5)
    os.unlink(tmp.name)


class NumericType(click.ParamType):
    name = "numeric"

    def convert(self, value, param, ctx):
        # strip commas if the got passed in a bbox
        value = value.replace(",", "")
        try:
            return float(value)
        except TypeError:
            self.fail(
                "expected string for int() conversion, got "
                f"{value!r} of type {type(value).__name__}",
                param,
                ctx,
            )
        except ValueError:
            self.fail(f"{value!r} is not a value that can be converted to a float", param, ctx)


NUM_TYPE = NumericType()


@click.command()
@click.option(
    "--acq-id",
    "acq_ids",
    multiple=True,
    help="Limit the Select to these acquisition IDs. Can be provided multiple times",
)
@click.option("--datetime", help="Limit the Select to a given date or date range")
@click.option(
    "--intersects",
    help="Search for tiles that intersect this geometry in WKT format, or load geometry from a file path",
)
@click.option(
    "--bbox",
    nargs=4,
    type=NUM_TYPE,
    help="like `intersects`, but limits search to a WGS84 bounding box in format `--bbox XMIN YMIN XMAX YMAX`",
)
@click.option(
    "--stack-depth",
    type=int,
    help="If provided, only return tiles where the stack depth can be fulfilled.",
)
@click.option(
    "--filter",
    nargs=3,
    multiple=True,
    help="Add a filter statement in the form of `--filter <property> <operator> <value>. See docs for full filter syntax.",
)
@click.option(
    "--image-age",
    "image_age_category",
    multiple=True,
    type=click.Choice(["fresh", "standard", "training"], case_sensitive=False),
    help="Limit imagery to an image age category. Can be used more than once if age ranges are contiguous.",
)
@click.option(
    "--min-cloud-free",
    type=NUM_TYPE,
    help="Shortcut to filter the minimum percentage of cloud-free cover. A value of 100 means an image must be 100% free of clouds.",
)
@click.option(
    "--min-data", type=NUM_TYPE, help="Shortcut to filter the minimum percentage of valid pixels."
)
@click.option(
    "--bba", is_flag=True, help="Restrict imagery to acquisitions that meet BBA requirements"
)
@click.option("--verbose", "-v", is_flag=True, help="Return more information about the Select")
def submit(**kwargs):
    """Submits a Select request to the server.

    There are numerous options and filter possibilities - consult the documentation for more information.

    At a minimum, you must supply at least one of the following:
    - a spatial filter - `intersects` or `bbox`
    - specific acquisition ID or IDs

    The Select service will try to compute results within 20 seconds. If it takes longer than 20 seconds the command
    will return a job ID that you can check for completion with `max-ard select status SELECT_ID`
    """
    # pop out kwargs that don't go the API
    min_cloud_free = kwargs.pop("min_cloud_free")
    min_data = kwargs.pop("min_data")
    filters = kwargs.pop("filter", [])
    for_bba = kwargs.pop("bba", False)
    verbose = kwargs.pop("verbose", False)

    query = defaultdict(dict)
    for prop, op, value in filters:
        # validate props and ops
        if op in ["eq", "ne", "gt", "gte", "lt", "lte"]:
            try:
                value = float(value)
            except:
                pass
        # turn comma-delimited strings to lists
        # and try to cast numbers if needed
        elif op == "in":
            value = value.split(",")
            value = [v.strip() for v in value]
            try:
                value = [float(v) for v in value]
            except:
                pass

        query[prop][op] = value

    # orders with BBA enabled will be rejected if the off-nadir angle of
    # any acquisition is greater than 30 degrees, so don't select images that
    # could cause a failure if the select is ordered
    if for_bba:
        query["view:off_nadir"]["lte"] = 30

    if min_cloud_free is not None:
        if min_cloud_free < 100:
            query["aoi:cloud_free_percentage"]["gte"] = min_cloud_free
        else:
            query["aoi:cloud_free_percentage"]["eq"] = 100

    if min_data is not None:
        if min_data < 100:
            query["aoi:data_percentage"]["gte"] = min_data
        else:
            query["aoi:data_percentage"]["eq"] = 100

    kwargs["query"] = query
    try:
        s = Select(**kwargs)
    except Exception as e:
        click.secho(f"There was an error in your parameters: {e}", fg="red", err=True)
        exit()

    try:
        s.submit()
    except BadARDRequest as e:
        click.secho(f"There was an error in your request: {e}", fg="red", err=True)
        if verbose:
            pp = pprint.PrettyPrinter(indent=4)
            click.secho("Request payload:", fg="red", err=True)
            click.secho(pp.pformat(s.request.to_payload()), fg="red", err=True)
        exit()
    if s.finished:
        click.secho(f"Select {s.select_id} has completed", fg="green")
    else:
        click.secho(f"Select {s.select_id} is still running", fg="cyan")
        click.secho(f"Run `max-ard select status {s.select_id}` to check the status", fg="cyan")
        click.secho('When the status is "SUCCEEDED" you can: ', fg="cyan")
    click.secho(
        f"Run `max-ard select describe {s.select_id}` to see a basic overview of the results",
        fg="cyan",
    )
    click.secho(
        f"Run `max-ard select browse {s.select_id}` to launch a browser-based map view of the results",
        fg="cyan",
    )
    click.secho("Run `max-ard select` to see all the `select` commands", fg="cyan")
    if verbose:
        pp = pprint.PrettyPrinter(indent=4)
        click.secho("Request payload:", fg="cyan")
        click.secho(pp.pformat(s.request.to_payload()), fg="cyan")


select.add_command(status)
select.add_command(describe)
select.add_command(browse)
select.add_command(submit)
select.add_command(url)



import os
from configparser import ConfigParser
from getpass import getpass

import click

from max_ard.admin import AccountManager
from max_ard.commands.grid import cell
from max_ard.commands.order import order
from max_ard.commands.select import select
from max_ard.commands.storage import storage
from max_ard.session import get_self, get_user_session, read_token, write_config


def check_access():
    get_self(get_user_session())
    return True


@click.group()
def max_ard():
    """Maxar ARD tools"""
    pass


@click.command()
@click.option("--username", help="ARD username")
@click.option("--password", help="ARD password")
def login(username, password):
    """Creates credential file using passed options --username and --password and tests access
    to the ARD system. If run without options,
    will only check if existing credential file at ~/.ard-config contains valid credentials"""

    if username or password:
        click.secho("... Creating config file ~/.ard-config", fg="cyan")
        write_config(username, password)
    else:
        if not os.path.exists(os.path.expanduser("~/.ard-config")):
            click.secho("No credential file exists - let's make a new one.")
            username = input("ARD account name? ")
            password = getpass(prompt="ARD account password? ")
            write_config(username, password)
    try:
        click.secho("... Checking access", fg="cyan")
        check_access()
        click.secho("ARD access confirmed", fg="green")
    except Exception as ex:
        click.secho(f"There was an error logging in: {ex}", fg="red", err=True)


@click.command()
@click.option("--reset", "-r", is_flag=True, help="Resets token (removes it from the config file)")
def token(reset):
    """ If run without options, returns the current token. 

        Handy for substitution in bash commands like curl:

            \b
            curl -H 'Accept: application/json' \\
                 -H "Authorization: Bearer $(max_ard token)" \\
                 https://hostname/api/myresource
    """

    if reset:
        config_file = os.path.expanduser("~/.ard-config")
        cfg = ConfigParser()
        cfg.read(config_file)
        cfg.remove_section("ard_token")
        click.secho("Token reset", fg="green")
        with open(config_file, "w") as sink:
            cfg.write(sink)
    else:
        check_access()
        click.secho(read_token(), fg="green")


@click.command()
def account():
    """Get basic account information for the current user"""

    self = get_self(get_user_session())
    user = self["user"]
    limits = user["limits"]

    click.secho(f'Account ID: {user["account_id"]}')
    click.secho(f'User ID: {user["user_id"]}')
    click.secho(f'User Name: {user["name"]}')
    click.secho(f'User Email: {user["email"]}')
    if any(limits.values()):
        click.secho(f"Usage Limits:")
        click.secho(f'  Area: {limits["annual_subscription_fee_limit"]} sq.km')
        click.secho(f'  Subscription: ${limits["fresh_imagery_fee_limit"]}')
        click.secho(f'  Fresh Imagery: ${limits["fresh_imagery_fee_limit"]}')
        click.secho(f'  Standard Imagery: ${limits["standard_imagery_fee_limit"]}')
        click.secho(f'  Training Imagery: ${limits["training_imagery_fee_limit"]}')
    else:
        click.secho(f"Usage Limits: None")
    click.secho("")


@click.command()
@click.option("--user", help="ARD user ID")
@click.option("--account", help="ARD account")
@click.option("--start-date", "start_date", help="Start date YYYY-MM-DD")
@click.option("--end-date", "end_date", help="End date YYYY-MM-DD")
def usage(user, account, start_date, end_date):
    """Get the usage for an account or a user"""

    self = get_self(get_user_session())
    while not any([user, account, start_date, end_date]):
        source = input("Get usage for an [a]ccount or [u]ser? ")
        if source.lower() not in ["a", "u"]:
            click.secho('Please press "a" for account or "u" for user', fg="red")
            continue
        if source.lower() == "a":
            account = input("Account number [current account]? ")
            if not account:
                account = self["user"]["account_id"]
        else:
            user = input("User ID [current user]? ")
            if not user:
                user = self["user"]["user_id"]
        start_date = input("Start date YYYY-MM-DD [none]? ")
        end_date = input("End date YYYY-MM-DD [none]? ")

    manager = AccountManager(account)

    try:
        if user:
            usage = manager.get_user_usage(user, start_date, end_date)
        else:
            usage = manager.get_account_usage(start_date, end_date)
    except Exception as e:
        click.secho(
            f"Error: {e}",
            err=True,
            fg="red",
        )
        exit(1)

    click.secho(str(usage), fg="cyan")


max_ard.add_command(login)
max_ard.add_command(token)
max_ard.add_command(usage)
max_ard.add_command(account)


max_ard.add_command(storage)
max_ard.add_command(cell)
max_ard.add_command(cell)
max_ard.add_command(select)
max_ard.add_command(order)

if __name__ == "__main__":
    max_ard()



"""Create ARD Selects and view results

Provides
--------
  1. Objects representing Selects
  2. Objects representing detailed results of a Select


Documentation
-------------

Select API Overview

- https://ard.maxar.com/docs/select-and-order/select-tiles/

Select SDK Tutorial

- https://ard.maxar.com/docs/sdk/sdk/selecting-ard/

API request and response examples

- https://ard.maxar.com/docs/api-reference/select/select_resource/

Example
-------

    >>> bbox = [-106.8, 35.1, -106.4, 35.4]
    >>> datetime =  "2020-07-01T00:00:00Z/2021-01-25T00:00:00Z"
    >>> query = {
            "platform": {
            "eq": "worldview-02"
            },
            "aoi:cloud_free_percentage": {
            "gte": 95
            },
            "aoi:data_percentage": {
            "gte": 75
            }
        }
    >>> select = Select(datetime=datetime, bbox=bbox, query=query, stack_depth=3)
    >>> select.submit()
    >>> s.state
    SUCCEEDED
    >>> s.select_id
    5629729628519955012
    >>> s.usage
    SelectUsage(
        area=UsageArea(fresh_imagery_sqkm=0.0, standard_imagery_sqkm=2317.0, training_imagery_sqkm=0.0, total_imagery_sqkm=2317.0, estimate=True), 
        cost=UsageCost(fresh_imagery_cost=0.0, standard_imagery_cost=24.0, training_imagery_cost=0.0, total_imagery_cost=24.0, estimate=True), 
        limits=UsageLimits(limit_sqkm=-1.0, fresh_imagery_fee_limit=-1.0, standard_imagery_fee_limit=-1.0, training_imagery_fee_limit=-1.0, annual_subscription_fee_limit=-1.0), 
        available=UsageAvailable(available_sqkm=-1.0, fresh_imagery_balance=-1.0, standard_imagery_balance=-1.0, training_imagery_balance=-1.0, total_imagery_balance=-1.0), 
        usage_as_of='2021-07-27T18:29:11Z')
    >>> results = s.results
    >>> print(results)
    <SelectResult (95 tiles in 12 acquisitions) >
    >>> results.dates
    ['2020-07-10',
    '2020-07-24',
    '2020-07-29',
    '2020-09-08',
    '2020-10-21',
    '2021-01-16']
 
"""
import json
import warnings
from functools import lru_cache, wraps
from math import cos, radians
from pathlib import Path
from time import sleep
from typing import List, Union

import requests
from maxar_ard_grid import Cell
from pydantic import BaseModel
from shapely.geometry import LineString, Point, Polygon, box, mapping, shape
from shapely.wkt import loads

from max_ard.admin import AdminUsage
from max_ard.base_collections import ARDModel, BaseCollection, Succeeded, Submitted
from max_ard.exceptions import (
    ARDServerException,
    BadARDRequest,
    NotFinished,
    NotSubmitted,
    SelectError,
)
from max_ard.io import KmlDoc, ShpDoc, convert_to_shapely
from max_ard.session import ard_url, get_user_session

__all__ = ("Select", "SelectResult", "SelectTile")


class SelectRequest(ARDModel):
    """An object to hold the request to the Select Service"""

    ids: List[str] = None
    datetime: str = None
    stack_depth: int = None
    intersects: Union[str, dict] = None
    bbox: Union[List, tuple] = None
    query: dict = {}
    image_age_category: Union[List, tuple] = None

    class Config:
        extra = "allow"
        validate_assignment = True


class SelectUsage(AdminUsage):
    """Model to hold usage information returned from a Select

    Inherits from AdminUsage, however this adds a 'usage_as_of' field so that Select products
    that show this information can indicate that the account balance was valid when the Select was run.

    """

    usage_as_of: str

    def __getitem__(self, key):
        """Get old usage values for backwards compatibility"""
        warnings.warn(
            "Usage objects are now Pydantic models with more fields, simple dict access will be deprecated",
            DeprecationWarning,
        )
        if key == "limit_sqkm":
            return self.limits.limit_sqkm
        elif key == "available_sqkm":
            return self.available.available_sqkm
        elif key == "usage_as_of":
            return self.usage_as_of
        elif key == "fresh_imagery_sqkm":
            return self.area.fresh_imagery_sqkm
        elif key == "standard_imagery_sqkm":
            return self.area.standard_imagery_sqkm
        elif key == "selection_sqkm":
            return self.area.total_imagery_sqkm
        elif key == "training_imagery_sqkm":
            return self.area.training_imagery_sqkm
        else:
            raise KeyError()


class SelectResponse(BaseModel):
    """an object to hold the Select API response"""

    id: str
    status: str
    request_details: dict = None
    stack_depth_summary: dict = None
    unique_acquisitions: int = None
    usage: SelectUsage = None
    links: dict = None
    error_message: dict = None

    class Config:
        extra = "allow"
        allow_mutation: False


class Select(Succeeded, Submitted):
    """An ARD API Select object

    Parameters
    ----------
    acq_ids : iterable of str
        An iterable of acquisition IDs to search for
    datetime : str
        Date or date range string
    intersects : Geometry-like objects, str (optional))
        Geometry to intersect, can be most geometry or spatial objects, or a path (see Notes)
    bbox : interable of numeric
        Bounding box in WGS84 coordinates, [west, south, east, north]
    query : dict
        Query dictionary
    stack_depth: int or None, optional
        Maximum number of tiles to return
    image_age_category : iterable of str, optional
        One or
    session : session object, optional
        A user_session or application_session object

    Attributes
    ----------
    session : session object or None
        A user_session or application_session object
    submitted : bool
        True if Select has been submitted via `max_ard.select.Select.submit`
    request : Pydantic model
        Parameters are loaded into a Pydantic model of the HTTP API request
    response : Pydantic model
        Pydantic model of the server response to API call. Status and submit calls
        return the same payload


    Notes
    -----

    Intersects inputs:
        Geometry objects: Shapely shapes, objects supporting __geo_interface__, geojson-like dicts,
            geojson and wkt strings
        Geometry iterables: iterables of above, Fiona readers
        File paths: most spatial file formats. WKT and Geojson supported with base install, other formats
            require Fiona for reading"""

    # store an authenticated Requests session in the class

    def __init__(
        self,
        acq_ids=None,
        datetime=None,
        intersects=None,
        bbox=None,
        query={},
        stack_depth=None,
        image_age_category=None,
        session=None,
    ):

        self.session = session or get_user_session()

        if intersects is not None:
            intersects = convert_to_shapely(intersects).wkt

        self.request = SelectRequest(
            ids=acq_ids,
            datetime=datetime,
            intersects=intersects,
            bbox=bbox,
            stack_depth=stack_depth,
            query=query,
            image_age_category=image_age_category,
        )
        self.response = None

    @property
    def submitted(self):
        if self.response:
            return self.response.id is not None
        else:
            return False

    @property
    @Submitted.required
    def running(self):
        """The Select is currently running"""

        return self.status == "RUNNING"

    @property
    @Submitted.required
    def finished(self):
        """The Select has finished running but may have failed"""

        return self.status != "RUNNING"

    @property
    @Submitted.required
    def succeeded(self):
        """The Select has finished running and has succeeded"""

        return self.status == "SUCCEEDED"

    @property
    @Submitted.required
    def failed(self):
        """The Select has finished running but has failed"""

        return self.status in ["FAILED", "ERROR"]

    @Submitted.required
    def wait_for_success(self, interval: int = 5) -> None:
        """Wait for the Select to succeed

        Parameters
        ----------
        interval: numeric, optional
            polling interval for success, default is 5 secs

        Raises
        ------
        SelectError
            An error in the selection caused it to fail"""

        while self.status == "RUNNING":
            sleep(interval)
        if self.status == "FAILED":
            error = self.response.error_message
            msg = f'{error["Error"]}: {error["Cause"]}'
            raise SelectError(msg)

    @property
    @Submitted.required
    def status(self):
        """Status of the select process: 'RUNNING', 'FINISHED', or 'FAILED'"""
        if self.response.status == "RUNNING":
            response = self.get_select(self.select_id, self.session)
            self.response = SelectResponse(**response)
        return self.response.status

    @property
    def state(self):
        """Alternate name for `max_ard.Select.status`, will be deprecated"""
        return self.status

    @property
    @Submitted.required
    def select_id(self):
        """ID of the Select"""

        return self.response.id

    @property
    @Succeeded.required
    def usage(self):
        """Dictionary of data usage metrics"""

        return self.response.usage

    @classmethod
    def from_id(cls, select_id: str, session=None):
        """Create a Select object from an ID

        Parameters
        ----------
        select_id: str
            Select ID to hydrate into a Select object
        session : Session object, optional
            Authenticated session, such as from get_client_session()

        Returns
        -------
        Select"""

        if not session:
            session = get_user_session()
        instance = cls()
        instance.session = session
        response = cls.get_select(select_id, session)
        instance.response = SelectResponse(**response)
        if "request_details" in response:
            instance.request = SelectRequest(**response["request_details"])
        return instance

    @classmethod
    def get_select(cls, select_id, session=None):
        """Fetch raw data about a Select from an ID

        Parameters
        ----------
        select_id : str
            Select ID to fetch metadata for
        session: Session object, optional
            Authenticated session, such as from get_client_session()

        Returns
        -------
        dict
            API data for the given Select"""

        if not session:
            session = get_user_session()

        r = session.get(ard_url("select", "request", select_id))
        return r.json()

    @classmethod
    def send_select(cls, payload, session=None):
        """Send a request to the Select API

        Parameters
        ----------
        payload : dict
            Select API request payload
        session : Session object, optional
            Authenticated session, such as from get_client_session()

        Returns
        -------
        dict
            API response data for the given Select"""

        if not session:
            session = get_user_session()
        r = session.post(ard_url("select"), json=payload)
        return r.json()

    def submit(self):
        """Submit this Select to the API"""

        response = self.send_select(self.request.to_payload(), self.session)
        if self.submitted:
            warnings.warn("The Select has already been submitted")
            return
        self.response = SelectResponse(**response)
        # update request in case defaults were applied
        self.request = SelectRequest(**self.response.request_details)

    @lru_cache()
    @Succeeded.required
    def get_link_contents(self, name):
        """Get the contents of an Select result file via its signed link

        Parameters
        ----------
        name : str
            The Select result file name

        Returns
        -------
        str
            Link contents"""

        temp_url = self.get_signed_link(name)
        # unauthenticated, don't send token
        r = requests.get(temp_url)
        r.raise_for_status()
        return r.text

    @Succeeded.required
    def get_signed_link(self, name):
        """Get the signed link for a Select result file

        Parameters
        ----------
        name : str
            The Select result file name

        Returns
        -------
        str
            signed URL"""

        url = self.response.links[name]
        r = self.session.get(url)
        r.raise_for_status()
        return r.json()["download_link"]

    @Succeeded.required
    def copy_file(self, name, dir="."):
        """Copy a Select result file to a local location

        Parameters
        ----------
        name : str
            The Select result file name
        dir : str, optional
            Local directory location to copy to, file will retain its name"""

        # TODO get the output filename
        path = Path(dir, f"{self.select_id}.{name}")

        with open(path, "w") as out:
            file = self.get_link_contents(name)
            out.write(file)

    @property
    @lru_cache()
    @Succeeded.required
    def results(self):
        """The results of a select converted to Python objects"""

        return SelectResult.from_geojson(json.loads(self.get_link_contents("geojson")))

    def __repr__(self) -> str:
        if not self.submitted:
            return f"<ARD Select (unsubmitted)>"
        elif self.succeeded:
            return f"<ARD Select {self.select_id}>"
        else:
            return f"<ARD Select ({self.status})>"


class SelectTile:
    """An ARD tile identified in material selection

    These are generated by the SelectResult hydration and would not be initalized independently

    Parameters
    ----------
    properties : dict
        The properties of the source tile

    Attributes
    ----------
    date : str
        Date of object as YYYY-MM-DD string
    acq_id : str
        Acquisition ID of tile source
    quadkey : str
        Quadkey of the tile
    zone : int
        UTM zone of the tile
    cell_id : str
        Cell ID, example Z13-031133320001"""

    def __init__(self, properties: dict):
        self.properties = properties
        self.date = self.properties["date"]
        self.acq_id = self.properties["acquisition_id"]
        self.quadkey = self.properties["tile:quadkey"]
        self.zone = int(self.properties["tile:zone"])
        self.cell_id = "Z%02d-%s" % (self.zone, self.quadkey)

    @property
    @lru_cache()
    def cell(self) -> Cell:
        """The ARD Grid cell of the tile"""
        return Cell(self.quadkey, zone=self.zone)

    @property
    @lru_cache()
    def data_mask(self):
        return loads(self.properties["wkt"])

    @property
    @lru_cache()
    def no_data_mask(self):
        return shape(self.cell).difference(self.data_mask)

    @property
    def thumbnail_url(self) -> str:
        """The URL thumbnail of the tile from the Browse imagery"""

        return ard_url("browse", "preview", self.acq_id, self.cell_id)

    @property
    @lru_cache()
    def __geo_interface__(self) -> dict:
        return mapping(self.data_mask)

    def __repr__(self) -> str:
        return f"<SelectTile of {self.acq_id} at {self.cell_id}>"


class SelectResult(BaseCollection):
    """The results of a Select or MetaSelect operation.

    This object is converted from the GeoJSON FeatureCollections
    returned by Selects and MetaSelects into Python objects"""

    def __init__(self):
        super().__init__()

    @classmethod
    def from_geojson(cls, geojson):
        self = cls()
        for feature in geojson["features"]:
            if "best_matches" in feature["properties"]:
                for match in feature["properties"]["best_matches"]:
                    self.add_tile(SelectTile(match))
            else:
                try:
                    self.add_tile(SelectTile(feature["properties"]))
                except (KeyError, TypeError):
                    pass
        return self

    def __repr__(self) -> str:
        return (
            f"<SelectResult ({len(self.tiles)} tiles in {len(self.acquisitions)} acquisitions) >"
        )


def ard_bbox_generator(input):
    """
    Takes a Shapely shape object and makes it into a valid intersect geometry for a Select based on a bounding box of the input object
    Args:
        input: a Shapely shape object (can be Point, MultiPoint, LineString, MultiLineString, Polygon, or MultiPolygon)
    Returns:
        Shapely polygon (rectangle) that is large enough to be a valid intersect geometry for a Select
    """

    # we set DEG_KM to 0.01 because at the equator 0.01° = 1.11 km for latitude or longitude and the API wants an area of at least 1 km^2
    # we want the latitude diff to be at least 0.01 but the longitude diff / cos(latitude diff) should equal at least 0.01
    DEG_KM = 0.01
    min_area = DEG_KM * DEG_KM

    try:  # make sure we have a valid Shape
        input.centroid
    except AttributeError:
        raise TypeError(
            "Input type must be a Shapely Point, MultiPoint, LineString, MultiLineString, Polygon, or MultiPolygon"
        )

    # make a bounding box for our input
    bbox = input.envelope

    xmin, ymin, xmax, ymax = bbox.bounds
    xdiff = xmax - xmin
    ydiff = ymax - ymin

    cosy = cos(radians(ymax))

    if bbox.area * cos(radians(ymax)) >= min_area:
        # if area of bbox is large enough, then go ahead and return it
        return bbox

    elif xdiff == 0 and ydiff == 0:
        # we have a point, so increase x and y in equal measure

        extra_x = (DEG_KM / cosy) * 0.5
        extra_y = DEG_KM * 0.5

        xmin -= extra_x
        xmax += extra_x

        ymin -= extra_y
        ymax += extra_y

    elif xdiff >= ydiff:
        # increase y-dimension
        ydim = min_area / (xdiff * cosy)
        extra_y = ydim / 2
        ymin -= extra_y
        ymax += extra_y

    else:  # ydiff > xdiff
        # increase x-dimension
        xdim = min_area / ydiff
        extra_x = xdim / 2
        xmin -= extra_x
        xmax += extra_x

    return box(xmin, ymin, xmax, ymax)



"""Perform admin functions

Provides
--------
  Administrative functions

  - Check data usage at the user or account level
  - add/delete/get/list stored user credentials

"""
import base64
import urllib

from datetime import datetime
from functools import wraps
from posixpath import join
from typing import Union

from pydantic import BaseModel

from max_ard.base_collections import hydrate, UsageArea, UsageCost, UsageLimits, UsageAvailable
from max_ard.exceptions import MissingARDResource
from max_ard.monitor import Monitor
from max_ard.session import ard_url, get_self, get_user_session, paginated_response


__all__ = ["AccountManager"]


class AdminUsage(BaseModel):
    """A model to hold account or user usage"""

    area: UsageArea
    cost: UsageCost
    limits: UsageLimits
    available: UsageAvailable

    def __str__(self):
        def u(val):
            if val == -1:
                return "unlimited"
            else:
                return f"{val:.2f}"

        return f"""
 Data Usage
 ----------

  Account limit: {u(self.limits.limit_sqkm)} sq.km (${u(self.limits.annual_subscription_fee_limit)})

  Imagery ordered: {self.area.total_imagery_sqkm} sq.km (${self.cost.total_imagery_cost})
  ├ Fresh (< 90 days): {self.area.fresh_imagery_sqkm} sq.km (${self.cost.fresh_imagery_cost})
  ├ Standard (90 days - 3 years): {self.area.standard_imagery_sqkm} sq.km (${self.cost.standard_imagery_cost})
  └ Training (> 3 years): {self.area.training_imagery_sqkm} sq.km (${self.cost.training_imagery_cost})

  Remaining balance: {u(self.available.available_sqkm)} sq.km (${u(self.available.total_imagery_balance)})
  
  """


class RegisteredCredentials(BaseModel):
    "A model to hold stored user credentials"
    credentials_id: str
    account_id: str
    description: Union[str, None]
    created: str
    modified: str

    def __repr__(self):
        if self.short_description not in [None, ""]:
            return f'<RegisteredCredentials "{self.credentials_id}" ({self.short_description})>'
        else:
            return f'<RegisteredCredentials "{self.credentials_id}">'

    @property
    def short_description(self):
        if self.description is None:
            return ""
        if len(self.description) > 20:
            return self.description[0:20] + "..."
        else:
            return self.description


class AccountManager:
    """Manages account-related actions

    Arguments
    ---------
    account_id : str (optional)
        Account ID to manage. If not provided, will use the current user's account
    session : Requests session object (optional)
        The session to use to communicate with the API. If not provided, uses the current user's authenticated session
    """

    __all__ = ["get_account_usage", "get_user_usage"]

    def __init__(self, account_id=None, session=None) -> None:
        if not session:
            session = get_user_session()
        self.session = session
        if not account_id:
            account_id = get_self(self.session)["user"]["account_id"]
        self.account_id = account_id

    def admin_url(self, *args):
        return ard_url("admin", "account", self.account_id, *args)

    @property
    def properties(self):
        return self.session.get(self.admin_url()).json()["account"]

    ########
    # Usage
    ########

    def _validate_dates(self, *args):
        """Validate that dates are YYYY-MM-DD

        Arguments
        ---------
        *args : str
            One or more date strings to validate

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If one or more dates was not YYYY-MM-DD"""

        for date in args:
            if date not in [None, ""]:
                try:
                    datetime.strptime(date, "%Y-%m-%d")
                except ValueError:
                    raise ValueError("Dates must be YYYY-MM-DD")

    def get_account_usage(self, start_date=None, end_date=None):
        """Get account-level data usage

        Parameters
        ----------
        start_date : str (optional)
            Start date to calculate usage
        start_date : str (optional)
            End date to calculate usage

        Returns
        -------
        AdminUsage
            Object model of data usage
        """

        url = ard_url("usage", "account", self.account_id)

        self._validate_dates(start_date, end_date)

        params = {"start_date": start_date, "end_date": end_date}

        r = self.session.get(url, params=params)
        return AdminUsage(**r.json()["usage"])

    def get_user_usage(self, username=None, start_date=None, end_date=None):
        """Get user-level data usage

        Parameters
        ----------
        username : str (optional)
            User name to calculate usage for, if not provided the current user is used
        start_date : str (optional)
            Start date to calculate usage
        end_date : str (optional)
            End date to calculate usage

        Returns
        -------
        AdminUsage
            Object model of data usage
        """

        if not username:
            username = get_self(self.session)["user"]["user_id"]

        url = ard_url("usage", "user", username)

        self._validate_dates(start_date, end_date)

        params = {"start_date": start_date, "end_date": end_date}

        r = self.session.get(url, params=params)
        return AdminUsage(**r.json()["usage"])

    #####
    # Credentials Storage
    #####

    def add_credentials(self, name, credentials, description=""):
        """Saves credentials in the Crendentials API

        The stored credentials key (SAS Url for Azure or Base64-encoded Credentials JSON for GCS) is never returned.
        Only the name, description, and other metadata about the credentials object is provided.

        This is only available to Admin users. However all users in the account may use a stored credential name
        to provide write access when ordering data.

        Parameters
        ----------
        name : str
            Name of the stored credentials, should have URL-friendly characters (no spaces)
        credentials : str
            Credentials secret to store (SAS Url for Azure or Base64 Credentials JSON for GCS).
            If this is a path to a GCS JSON file, it will be read and Base64 encoded.
        description : str (optional)
            Description of the credentials

        Returns
        -------
        dict
            API response JSON dictionary
        """

        if urllib.parse.quote(name) != name:
            raise ValueError(
                "Stored credential names should not use characters that require URL-encoding"
            )
        url = self.admin_url("credentials", name)

        try:
            with open(credentials, "rb") as binary_file:
                encoded = base64.b64encode(binary_file.read())
                credentials = encoded.decode("ascii")

        except (FileNotFoundError, OSError):
            pass

        payload = {"credentials": credentials, "description": description}
        response = self.session.put(url, json=payload)

        if response.status_code == 204:
            return None
        else:
            return response.json()

    def get_credentials(self, name, raw=False):
        """Retrieves information about stored credentials.

        The stored credentials key (SAS Url for Azure or Crendentials JSON for GCS) is never returned,
        only the name, description, and other metadata about the credentials object.

        This is only available to Admin users. However all users in the account may use a stored credential name
        to provide write access when ordering data.

        Parameters
        ----------
        name : str
            Name of the stored credentials
        raw : bool (optional)
            Return the response JSON

        Returns
        -------
        RegisteredCredentials
            Object model of credentials
        """

        url = self.admin_url("credentials", name)
        response = self.session.get(url)
        if raw:
            return response.json()
        else:
            return RegisteredCredentials(**response.json()["registered_credentials"])

    def list_credentials(self, raw=False):
        """Retrieves all stored credentials in an account.

        The stored credentials key (SAS Url for Azure or Crendentials JSON for GCS) is never returned,
        only the name, description, and other metadata about the credentials object.

        This is only available to Admin users. However all users in the account may use a stored credential name
        to provide write access when ordering data.

        Parameters
        ----------
        raw : bool (optional)
            Return the response JSON

        Returns
        -------
        List of RegisteredCredentials
            List of object models of credentials
        """

        # responses can be paginated
        id_key = lambda credential: credential["registered_credentials"]["credentials_id"]
        url = self.admin_url("credentials")

        credentials = paginated_response(self.session, url, id_key)

        if raw:
            return credentials
        else:
            object_key = lambda credential: credential["registered_credentials"]
            return hydrate(RegisteredCredentials, credentials, key=object_key)

    def delete_credentials(self, name):
        """Deletes stored credentials by name.

        This is only available to Admin users. However all users in the account may use a stored credential name
        to provide write access when ordering data.

        Parameters
        ----------
        name : str
            Name of the credential object

        """
        url = self.admin_url("credentials", name)
        return self.session.delete(url)

    #####
    # Monitors
    #####

    def list_monitors(self, raw=False):
        """Retrieves all monitors for an account.

        Parameters
        ----------
        raw : bool (optional)
            Return the response JSON

        Returns
        -------
        List of Monitors
            List of object models of monitors
        """

        id_key = lambda monitor: monitor["monitor_id"]

        url = self.admin_url("monitor")
        monitors = paginated_response(self.session, url, id_key)

        if raw:
            return monitors
        else:
            return [Monitor.from_response(monitor) for monitor in monitors]



"""Collection objects representing stored ARD data

Provides
--------
  1. Connect to local or cloud ARD tiles
  2. Filter collections of tiles
  3. Furter iterate over collection of tiles spatially or temporally
  4. Pythonic access of ARD images and masks

Reference
---------

SDK Ordering Tutorial

- https://ard.maxar.com/docs/sdk/sdk/ordering-ard/


Example
-------
>>> collection = ARDCollection('bucket/prefix')
>>> for tile in collection.tiles:
        print tile.cell.bounds

"""

import json
import os
import re
import warnings
from datetime import date, datetime
from functools import lru_cache
from pathlib import Path, PurePosixPath
from posixpath import join as urljoin
from typing import Any, Iterable, List, Optional, Union
from urllib.parse import urlparse

from fsspec import filesystem
from fsspec.implementations.local import LocalFileSystem
from fsspec.spec import AbstractFileSystem
from maxar_ard_grid import Cell, covers
from shapely.geometry import GeometryCollection, box, mapping, shape

from max_ard.base_collections import BaseCollection
from max_ard.dependency_support import HAS_FIONA, HAS_RASTERIO
from max_ard.exceptions import MissingDependency, NotFinished
from max_ard.order import Order
from max_ard.processing import COGReader, read_windows, write_windows

# Optional dependencies


if HAS_RASTERIO:
    import rasterio
if HAS_FIONA:
    import fiona


class ARDTile:
    """Represents an ARD Tile - the images and vectors from one acquistion in one grid cell

    These objects have dynamic accessors for assets - calling the asset name (with underscores substituted
    for dashes) is the same as calling `ARDTile.open_asset(<name>)`.

    Example: `ARDTile.visual` returns a Rasterio dataset reader of the `visual` raster asset.

    Vector assets can also be accessed "inverted" by prefixing `no_` to the asset name.

    Example: `ARDTile.no_cloud_mask` will return the geometry of areas *not* covered by clouds.
    """

    # matching regexes
    # don't be a smartass and combine them

    # matches quadkeys
    CELL_REGEX = re.compile(r"[/-](\d{2})[-/]([0123]{12})[/-]")
    # matches regular catalog IDs
    CAT_ID_REGEX = re.compile(r"10[1-6][0AC][0-9A-F]{3}0[0-9A-F]{6}00", re.I)
    # matches WV4 catIDs (UUID v4)
    WV4_REGEX = re.compile(
        r"([0-9a-f]{8}-[0-9a-f]{4}-4[0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}-inv)", re.I
    )
    # matches dates yyyy-mm-dd
    DATE_REGEX = re.compile(r"\d{4}-\d{2}-\d{2}")

    def __init__(self) -> None:
        self.date = None  #: date of cell

        #: acquisition ID
        self.acq_id = None
        self.quadkey = None
        """ ARD Grid quadkey of the tile's cell """
        self.zone = None
        """ UTM zone of the tile """

    @property
    def properties(self) -> dict:
        """metadata of the ARD tile"""
        return self.stac_item["properties"]

    def __getattr__(self, name: str) -> Any:
        if name == "clouds":
            warnings.warn("clouds attribute should be accessed with ARDTile.cloud_mask")
            return self.open_asset("cloud-mask")
        elif name == "extent":
            warnings.warn("extent attribute should be accessed with ARDTile.data_mask")
            return self.open_asset("data-mask")
        elif name == "nodata":
            warnings.warn("clouds attribute should be accessed with ARDTile.no_data_mask")
            return shape(self.cell).difference(self.data_mask)
        elif name.startswith("no_"):
            asset_name = name.split("_", 1)[1].replace("_", "-")
            source = getattr(self, asset_name)
            try:
                source.area
            except:
                raise AttributeError(
                    f"Attribute {name} is not spatial so it can't be flipped with 'no_'"
                )
            if name == "no_data_mask":
                return box(*self.visual.bounds).difference(source)
            else:
                return self.data_mask.difference(source)
        else:
            # workaround for the analytic assets that are underscored already
            if name == "ms_analytic" or name == "pan_analytic":
                asset_name = name
            else:
                asset_name = name.replace("_", "-")
            if asset_name in self.asset_paths:
                return self.open_asset(asset_name)
            else:
                raise AttributeError(f"Attribute .{name} not found")

    @property
    @lru_cache()
    def cell(self) -> Cell:
        """The ARD Grid cell of the tile"""
        return Cell(self.quadkey, zone=self.zone)

    @classmethod
    def from_doc(cls, fs: AbstractFileSystem, path: str):
        """
        Create an ARDTile object from a path to any of a tile's files.

        Parameters
        ----------
        fs : fsspec.AbstractFileSystem
           a `filesystem` object used to read files.
        path : str
            a path to any file in a tile.

        Returns
        -------
        ARDTile
        """

        tile = cls()
        tile.fs = fs

        # zone and quadkey
        try:
            matches = ARDTile.CELL_REGEX.search(path)
            tile.zone = matches[1]
            tile.quadkey = matches[2]

            # cat ID
            matches = ARDTile.CAT_ID_REGEX.search(path)
            if matches is None:
                matches = ARDTile.WV4_REGEX.search(path)
            tile.acq_id = matches[0]

            # date
            matches = ARDTile.DATE_REGEX.search(path)
            tile.date = matches[0]

        except IndexError:
            # use index error to find when we can't match a needed field
            raise ValueError("Non-conforming data found, please contact ard-support@maxar.com")

        # prefixes for opening files with gdal/ogr

        if "s3" in tile.fs.protocol:
            tile.gdal_prefix = f"/vsis3/"
        elif "gs" in tile.fs.protocol:
            tile.gdal_prefix = f"/vsigs/"
        elif "abfs" in tile.fs.protocol:
            tile.gdal_prefix = f"/vsiaz/"
        else:
            tile.gdal_prefix = ""

        # rebuild the path from properties
        # might be overkill
        base_prefix = path.split(f"/{tile.zone}/{tile.quadkey}")[0]
        base_prefix = urljoin(base_prefix, tile.zone, tile.quadkey)
        # legacy canvas
        if tile.date is not None:
            base_prefix = urljoin(base_prefix, tile.date)
        else:
            warnings.warn(
                "Legacy Canvas file path without date - this will not be supported in the future"
            )
        tile.base_prefix = base_prefix

        return tile

    @property
    @lru_cache()
    def stac_item(self) -> dict:
        item = f"{self.base_prefix}/{self.acq_id}.json"
        with self.fs.open(item) as f:
            return json.load(f)

    class AssetPaths(dict):
        """A dictionary that warns when old asset names are called
        The names will get deprecated at some point"""

        def __getitem__(self, key):
            if key == "clouds":
                warnings.warn(
                    'The "clouds" asset has been renamed to "cloud-mask", please update your call'
                )
            if key == "extent":
                warnings.warn(
                    'The "extent" asset has been renamed to "data-mask", please update your call'
                )
            return dict.__getitem__(self, key)

    @property
    @lru_cache()
    def asset_paths(self) -> AssetPaths:
        # TODO set a way to specify this so we don't have to check the stac every time
        # given a homogenous dataset

        assets = ARDTile.AssetPaths()

        # Item isn't a true STAC asset but we treat it as an asset of the ARDTile
        # we could also treat this as "self" but that's not as clear
        assets["item"] = f"{self.base_prefix}/{self.acq_id}.json"

        # once STAC sample data is fixed, we can pull the assets from the Item
        # then we can add new assets without updating the SDK
        stac_assets = self.stac_item["assets"]

        # old non-relative links
        if stac_assets["visual"]["href"].startswith("s3://"):
            assets["visual"] = f"{self.base_prefix}/{self.acq_id}-visual.tif"
            assets["pan"] = f"{self.base_prefix}/{self.acq_id}-pan.tif"
            assets["ms"] = f"{self.base_prefix}/{self.acq_id}-ms.tif"
            assets["cloud-mask"] = f"{self.base_prefix}/{self.acq_id}-cloud-mask.gpkg"
            assets["data-mask"] = f"{self.base_prefix}/{self.acq_id}-data-mask.gpkg"
        else:
            # use stac asset hrefs
            for name, asset in stac_assets.items():
                if type(self.fs) == LocalFileSystem:
                    assets[name] = str(Path(self.base_prefix) / Path(asset["href"]))
                else:
                    assets[name] = str(PurePosixPath(self.base_prefix) / Path(asset["href"]))

        # legacy naming
        assets["clouds"] = assets["cloud-mask"]
        assets["extent"] = assets["data-mask"]
        return assets

    def __repr__(self):
        return f"<ARDTile of {self.acq_id} at z{self.zone}-{self.quadkey}>"

    # there's a small speed boost plus some cost savings with GETs by caching
    # asset fetching
    @lru_cache(maxsize=16)
    def open_asset(self, name):
        """
        Open an asset by name, where `name` is a key in the STAC Item assets

        For information on the asset files see https://ard.maxar.com/docs/ard_order_delivery/about_ard_order_delivery/

        To open the tile STAC Item itself, use 'item'.
        Current values for ARD V3 are:

        STAC metadata:
        - item

        Vectors:
        - cloud-mask
        - cloud-shadow-mask
        - terrain-shadow-mask
        - ms-saturation-mask
        - data-mask

        Rasters:
        - visual
        - pan
        - ms
        - clouds
        - ms-saturation
        - terrain-shadows
        - water

        Vectors returned will be a single Shapely geometry object, a GeometryCollection of multiple objects, or if
        the mask is empty an empty GeometryCollection.

        Parameters
        ----------
        name : str
            name of asset to open, see list above

        Returns
        -------
        rasterio.DataSetReader
            for raster assets or
        shapely.geometry
            for vector assets or
        dict
            for STAC Items (JSON)
        """

        asset = self.asset_paths[name]
        if asset.endswith("tif"):
            if not HAS_RASTERIO:
                raise MissingDependency("Rasterio is required to read ARD rasters")
            # return rasterio.open(f'{self.gdal_prefix}{asset}')
            return COGReader(f"{self.gdal_prefix}{asset}")

        if asset.endswith("geojson") or asset.endswith("gpkg"):
            if not HAS_FIONA:
                raise MissingDependency("Fiona is required to access ARD masks as geometries")
            with fiona.open(f"{self.gdal_prefix}{asset}") as layer:
                geom = [shape(f["geometry"]) for f in layer]
            if len(geom) == 1:
                return geom[0]
            elif len(geom) == 0:
                return GeometryCollection()
            else:
                return GeometryCollection(geom)

        if asset.endswith("json"):
            with self.fs.open(asset) as f:
                return json.load(f)

    @property
    def __geo_interface__(self):
        """Python Geospatial Interface of tile's valid pixels"""
        return mapping(self.data_mask)


class ARDCollection(BaseCollection):
    """ARDCollections represent collections of S3 tiles. Currently the tiles
    can be stored in S3 or locally.

    Parameters
    ----------
    path : str
        Path to S3 prefix or STAC collection.
    aoi : shapely.geometry or str, optional
        Limit to finding tiles that cover this AOI, can be shapely geometry or most textual representations.
    acq_id_in : iterable of str, optional
        Limit to finding tiles from these acquisitions.
    zone_in : iterable of int, optional
        Limit to finding tiles in these zones.
    earliest_date : str or datetime.date or datetime.datetime, optional
        Limit to finding tiles after this date (strings must be YYYY-MM-DD).
    latest_date : str or datetime.date or datetime.datetime, optional
        Limit or finding tiles before this date.
    profile : str, optional
        AWS Profile to use when tiles are in S3.
    public : bool
        Access cloud data without authentication (for public buckets).


    The following parameters are also settable attributes and will trigger a rescan

    Attributes
    ----------
    path
    aoi
    acq_id_in
    zone_in
    earliest_date
    latest_date
    """

    def __init__(
        self,
        path: str,
        aoi: Optional[Any] = None,
        acq_id_in: Optional[Iterable[str]] = None,
        zone_in: Optional[Iterable[int]] = None,
        earliest_date: Optional[Union[str, datetime, date]] = None,
        latest_date: Optional[Union[str, datetime, date]] = None,
        profile: Optional[str] = None,
        public: Optional[bool] = False,
        **kwargs,
    ) -> None:

        self._dirty = True
        self._updating = False
        super().__init__()

        # TODO we may want to normalize AOIs to WGS84
        # but right now `covers` is probably capable enough
        self.aoi = aoi

        # disable authentication for public buckets
        # 'anon' is what fsspec calls it but 'public' makes more sense
        # however 'anon' can also have implications for Azure so
        # we should be able to override it just in case

        if "anon" in kwargs:
            anon = kwargs["anon"]
        else:
            # using the recommend Azure connection string access, anon needs to be True
            if path.startswith("az"):
                anon = True
            else:
                anon = public
        # For GDAL S3 locations we can turn off signing so
        # you can have expired credentials
        if anon and path.startswith("s3"):
            os.environ["AWS_NO_SIGN_REQUEST"] = "YES"

        # validate some inputs that have been problematic in the past
        assert zone_in is None or all(type(z) == int for z in zone_in), "Zones must be integers"
        assert zone_in is None or all(z - 1 in range(60) for z in zone_in), "Invalid zone numbers"
        if not zone_in:
            zone_in = None
        self.zone_in = zone_in

        assert acq_id_in is None or all(
            type(c) == str for c in acq_id_in
        ), "Catalog IDs must be strings"
        if not acq_id_in:
            acq_id_in = None
        self.acq_id_in = acq_id_in

        # store dates as strings, reformat if needed
        def format_date(d):
            if d is None:
                return None
            try:
                return d.strftime("%Y-%m-%d")
            except:
                assert re.match(
                    r"\d{4}-\d{2}-\d{2}", d
                ), "Dates must be YYYY-MM-DD strings, or date/datetime objects"
                return d

        self.earliest_date = format_date(earliest_date)
        self.latest_date = format_date(latest_date)

        # Set up the path and initialize the filesystem source
        self.path = path

        # this might need to be smarter (os.path?) for windows slashes
        if self.path[-1] == "/":
            self.path = self.path[:-1]
        if os.path.exists(self.path):
            self.storage_type = "file"
            self.fs_path = os.path.abspath(self.path)
        else:
            parsed = urlparse(self.path)
            if parsed.scheme == "":
                raise ValueError("Local path does not exist")
            if parsed.scheme not in ["s3", "gs", "az"]:
                raise ValueError("Unrecognized protocol (use s3://, gs://, or az://")
            self.storage_type = parsed.scheme
            # might not need this, s3fs doesn't care about paths leading with protocols
            # need to check gdal, etc
            self.fs_path = parsed.netloc + parsed.path

        self.fs = filesystem(self.storage_type, anon=anon, profile=profile)

        try:
            self.fs.ls(self.path)
        except:
            raise ValueError("Access error: check your path for errors and access permissions")

    def __setattr__(self, name, value):
        """Some attributes are read-only properties
        Related setters need to reset the collection state"""

        if name == "acq_ids":
            raise ValueError(".acq_ids is read-only - set .acq_id_in instead")
        if name == "zones":
            raise ValueError(".zones is read-only - set .zone_in instead")
        if name == "start_date":
            raise ValueError(".start_date is read-only - set .earliest_date instead")
        if name == "acq_ids":
            raise ValueError(".end_date is read-only - set .latest_date instead")
        if name in ["acq_id_in", "zone_in", "earliest_date", "latest_date", "aoi"]:
            # resets the bins
            self._dirty = True
        object.__setattr__(self, name, value)

    def __getattribute__(self, name):
        dirty = object.__getattribute__(self, "_dirty")
        updating = object.__getattribute__(self, "_updating")
        if (
            dirty
            and not updating
            and name
            in [
                "tiles",
                "acquisitions",
                "acquisition_ids",
                "stacks",
                "cells",
                "get_stack",
                "get_acquisition",
                "dates",
                "earliest_date",
                "latest_date",
                "zones",
                "read_windows",
                "write_windows",
            ]
        ):
            object.__setattr__(self, "_updating", True)
            self._scan_files()
            object.__setattr__(self, "_dirty", False)
            object.__setattr__(self, "_updating", False)
        return object.__getattribute__(self, name)

    def __repr__(self) -> str:
        return f"<ARDCollection at {self.path}/>"

    def _scan_files(self) -> None:
        self._reset()

        if self.aoi is not None:
            cells = set([f"{c.zone}/{c.quadkey}" for c in covers(self.aoi)])
        else:
            cells = []

        # STAC source
        if self.path.endswith("json"):

            with self.fs.open(self.path) as f:
                doc = json.load(f)

            # STAC Item
            if "type" in doc.keys():
                items = [self.path]

            # STAC Collection
            else:
                root_path = self.fs_path.split("order_collections")[0]
                items = []
                for link in doc["links"]:
                    if link["rel"] != "child":
                        continue
                    path = root_path + link["href"].replace("../", "")
                    with self.fs.open(path) as f:
                        links = json.load(f)["links"]
                        for link in links:
                            if link["rel"] == "item":
                                path = root_path + link["href"].replace("../", "")
                                items.append(path)

        # Filesystem source
        else:
            # build leading glob pattern based on zone & quadkey
            if self.aoi is not None:
                # shard on first 5 digits on quadkey
                # for parallel fetches
                qkbs = set()
                for qk in cells:
                    qkbs.add(qk[:5])
                paths = [f"{k}*/*/*.json" for k in qkbs]
            else:
                if not self.zone_in or len(self.zone_in) == 0:
                    paths = ["*/*/*/*.json"]
                else:
                    paths = [f"{z}/*/*/*.json" for z in self.zone_in]
            items = []
            for path in paths:
                full_path = f"{self.path}/{path}"
                for item in self.fs.glob(full_path):
                    items.append(item)

        # Filter out items
        for item in items:
            tile = ARDTile.from_doc(self.fs, item)
            if self.aoi is not None:
                if f"{tile.zone}/{tile.quadkey}" not in cells:
                    continue
            if self.acq_id_in is not None and tile.acq_id not in self.acq_id_in:
                continue
            if self.earliest_date is not None:
                if tile.date < self.earliest_date:
                    continue
            if self.latest_date is not None:
                if tile.date > self.latest_date:
                    continue
            self.add_tile(tile)

    def clear_filesystem_cache(self):
        """
        Clear the local cache of a remote filesystem

        Remote file systems (S3, Azure, Google Cloud) cache files locally for speed.
        If the remote files have changed while using an ARDCollectino, you can clear
        the cached files so that new files will be loaded.

        Parameters
        ----------

        Returns
        -------
        """
        self.fs.clear_instance_cache()

    @classmethod
    def from_order(cls, order, **kwargs):
        """
        Create an ARDCollection from an order ID.

        Accepts all filter keywords as used by class initialization.

        Parameters
        ----------
        order_id : str or Order
            Order object or Order ID to open.
        **kwargs
            Filter keywords as used by class initialization.

        Returns
        -------
        ARDCollection
        """

        if type(order) == str:
            order = Order.from_id(order)
        if not order.finished:
            raise NotFinished

        output_config = order.response.order.output_config
        if "bucket" in output_config:
            bucket = output_config["bucket"]
            prefix = output_config["prefix"]
            protocol = "s3"
        elif "amazon_s3" in output_config:
            bucket = output_config["amazon_s3"]["bucket"]
            prefix = output_config["amazon_s3"]["prefix"]
            protocol = "s3"
        elif "google_cloud_storage" in order:
            bucket = output_config["google_cloud_storage"]["bucket"]
            prefix = output_config["google_cloud_storage"]["prefix"]
            protocol = "gs"
        elif "azure_blob_storage" in order:
            bucket = output_config["azure_blob_storage"]["container"]
            prefix = output_config["azure_blob_storage"]["prefix"]
            protocol = "az"
        path = f"{protocol}://{bucket}/{prefix}/order_collections/{order.order_id}_root_collection.json"
        self = cls(path, **kwargs)
        return self

    def read_windows(self, *args, **kwargs):
        """See `max_ard.processing.read_windows`"""
        return read_windows(self, *args, **kwargs)

    def write_windows(self, *args, **kwargs):
        """See `max_ard.processing.write_windows`"""
        return write_windows(self, *args, **kwargs)


def copy(
    collection: ARDCollection, destination: str, flatten: bool = False, dry_run: bool = False
) -> None:
    """Copy data from a collection to a local location

    Parameters
    ----------
    collection : ARDCollection
        The ARDCollection to copy locally.
    destination : str
        Local path to copy to.
    flatten : bool False
        Write files in flat format without subdirectories.
    dry_run : bool False
        Print the source and destination of files instead of copying.

    Returns
    -------
    None
    """

    Path(destination).mkdir(parents=True, exist_ok=True)
    for tile in collection.tiles:
        paths = tile.asset_paths.copy()
        # legacy names
        try:
            del paths["clouds"]
            del paths["extent"]
        except:
            pass
        for path in paths.values():
            source = path
            target = path.replace(collection.fs_path, "")[1:]
            if flatten:
                target = urljoin(destination, target.replace("/", "-"))
            else:
                target = urljoin(destination, target)
            if dry_run:
                print(source, target)
            else:
                if not flatten:
                    Path(target).parent.mkdir(parents=True, exist_ok=True)
                collection.fs.get(source, target)



"""Place ARD Orders

Provides
--------
  1. Place an ARD order
  2. Test an ARD order with "dry-run" mode
  3. Check order status

Reference
---------

Ordering API overview

- https://ard.maxar.com/docs/select-and-order/order-ard/

SDK Ordering Tutorial

- https://ard.maxar.com/docs/sdk/sdk/ordering-ard/

API request and response examples

- https://ard.maxar.com/docs/api-reference/order/order_resource/


Example
-------

>>> from max_ard import Order
>>> order = Order(destination='my_bucket/my_ard_tiles', select_id=<Select ID>)
>>> order.add_email_notification('me@email.com')
>>> order.submit()

"""

import base64
import json
import warnings
from typing import List, Union

from pydantic import BaseModel

from max_ard.base_collections import (
    ARDModel,
    Submitted,
    UsageArea,
    UsageCost,
    EmailNotification,
    SNSNotification,
    hydrate_with_responses,
)
from max_ard.exceptions import ARDServerException, BadARDRequest, NotSubmitted
from max_ard.io import convert_to_shapely
from max_ard.session import ard_url, get_user_session, paginated_response


__all__ = ("Order",)

#######
# Models
#######


class Acquisition(BaseModel):
    id: str
    cells: List[str] = []

    class Config:
        validate_assignment = True


class Output_Config(BaseModel):
    bucket: str
    prefix: str
    role_arn: str = None

    class Config:
        validate_assignment = True


class S3_Config(BaseModel):
    sas_url: str = None
    bucket: str
    prefix: str

    class Config:
        validate_assignment = True
        extra: "forbid"


class GCS_Config(BaseModel):
    service_credentials: str = None
    credentials_id: str = None
    bucket: str
    prefix: str

    class Config:
        validate_assignment = True
        extra: "forbid"


class AZ_Config(BaseModel):
    sas_url: str = None
    credentials_id: str = None
    container: str
    prefix: str

    class Config:
        validate_assignment = True
        extra: "forbid"


class OrderRequest(ARDModel):
    acquisitions: List[Acquisition] = None
    select_id: str = None
    dry_run: bool = False
    # filter
    intersects: str = None
    bbox: List = None
    # destination
    output_config: dict = None
    settings: dict = {}
    notifications: List[Union[EmailNotification, SNSNotification]] = []
    metadata: dict = {}

    class Config:
        extra = "allow"
        validate_assignment = True


class OrderUsage(BaseModel):
    area: UsageArea = None
    cost: UsageCost = None


class OrderDetails(BaseModel):
    usage: OrderUsage = None

    class Config:
        extra = "allow"


class OrderResponse(BaseModel):
    id: str
    status: str
    status_message: str = None
    order: OrderDetails = None

    class Config:
        extra = "allow"
        validate_assignment = True


#####
# Classes
#####


class Order(Submitted):
    """An ARD API Order object

    Parameters
    ----------
    acquisitions : iterable of str or dict, optional
        An iterable of acquisitions to order, see Notes
    select_id : str
        A Select ID to order, see Notes
    destination:
        For S3 locations only, a path to storage location such as s3://my-bucket/my-prefix
    output_config:
        If not using `destination`, an output configuration dictionary, see API documentation for examples
    intersects : Geometry-like objects, str (optional))
        Geometry to intersect, can be most geometry or spatial objects, or a path (see Notes)
    bbox : interable of numeric, optional
        Like `intersects`, a bounding box in WGS84 coordinates, [XMIN YMIN XMAX YMAX]
    role_arn : str, optional
        A trusted Role ARN for the writer to assume so it can write tiles.
        This is not used if the s3 bucket policy allows writing tiles to the bucket.
    dry_run : bool, optional
        When true, runs pre-order checks to check if order is valid but does not generate imagery
    bba : bool, optional
        When true, Block Bundle Adjustment will be applied to the order.
    metadata : dict, optional
        User-supplied metadata
    settings : dict, optional
        Dictionary of settings to override outputs, see Notes
    session : session object, optional
        A user_session or application_session object

    Attributes
    ----------
    session : session object or None
        A user_session or application_session object
    dry_run : bool
        as above
    submitted : bool
        True if Select has been submitted via `max_ard.order.Order.submit`
    request : Pydantic model
        Parameters are loaded into a Pydantic model of the HTTP API request
    response : Pydantic model
        Pydantic model of the server response to API call. Status and submit calls
        return the same payload.

    Notes
    -----
    An order must specify `acquisitions` or `select_id`.

    If ordering by select ID, do not include acquisition IDs or an AOI in the order request.

    Acquisitions can be a list of acqusitions IDs, or a a list of dictionaries, with the keys
    `id` and `cells`. If no cells are specified, it is assumed all cells are wanted
    (subject to clipping by an AOI or BBOX):


      acquisitions=["103001009E8G3C90"]

      or

      acquisitions=[
        {
            "id": "103001007B478000",
            "cells": ["Z17-031313123113", "Z17-031313123112"]
        },
        {
            "id": "103001009E8C7C00",
            "cells": ["Z17-031313123113"]
        },
        {
            "id": "103001009E8G3C90"
        }
     ]

    Intersects inputs:
        Geometry objects: Shapely shapes, objects supporting __geo_interface__, geojson-like dicts,
            geojson and wkt strings
        Geometry iterables: iterables of above, Fiona readers
        File paths: most spatial file formats. WKT and Geojson supported with base install, other formats
            require Fiona for reading



    Settings dictionary defaults:
        You can override one or more of the following values:

        {
            "bundle_adjust": false,
            "cloud_mask": true,
            "data_mask": true,
            "healthy_vegetation_mask": true,
            "ms_analytic": true,
            "ms_saturation_mask": true,
            "pan_analytic": true,
            "pan_flare_mask": true,
            "terrain_shadow_mask": true,
            "visual": true,
            "water_mask": true
        }"""

    def __init__(
        self,
        acquisitions=None,
        select_id=None,
        destination=None,
        output_config=None,
        settings={},
        intersects=None,
        bbox=None,
        role_arn=None,
        dry_run=False,
        bba=False,
        metadata={},
        session=None,
    ):

        self.session = session or get_user_session()

        # check booleans aren't strings
        if type(dry_run) is not bool:
            raise ValueError(f"Parameter `dry_run` needs to be a boolean, not {type(dry_run)}")
        if type(bba) is not bool:
            raise ValueError(f"Parameter `bba` needs to be a boolean, not {type(bba)}")

        # bba to be moved to `settings` instead of top level param
        if "bundle_adjust" not in settings:
            settings["bundle_adjust"] = bba

        # set up the output config
        if destination is not None:
            destination = destination.replace("s3://", "")
            parts = destination.split("/", 1)
            bucket = parts[0]
            try:
                prefix = parts[1]
            except IndexError:
                prefix = ""
            # old style output config
            destination = {"bucket": bucket, "prefix": prefix, "role_arn": role_arn}
            output_config = self._validate_output({"destination": destination})

        else:
            if output_config is not None:
                output_config = self._validate_output(output_config)

        # set up acquisitions
        if type(acquisitions) == str:
            acquisitions = [Acquisition(id=acquisitions)]
        elif type(acquisitions) == list or type(acquisitions) == tuple:
            if len(acquisitions) > 0:
                if type(acquisitions[0]) == str:
                    acquisitions = [Acquisition(id=v) for v in acquisitions]
        elif type(acquisitions) == dict:
            acquisitions = [Acquisition(id=k, cells=v) for k, v in acquisitions.items()]

        # if the intersects is a Shapely geom, convert it to wkt
        if intersects is not None:
            intersects = convert_to_shapely(intersects).wkt

        self.request = OrderRequest(
            acquisitions=acquisitions,
            select_id=select_id,
            intersects=intersects,
            bbox=bbox,
            output_config=output_config,
            settings=settings,
            dry_run=dry_run,
            metadata=metadata,
        )
        self.dry_run = dry_run
        self.response = None

    def _validate_output(self, output_config):

        platforms = {
            "azure_blob_storage": AZ_Config,
            "google_cloud_storage": GCS_Config,
            "amazon_s3": S3_Config,
            "destination": Output_Config,
        }
        platform = list(output_config.keys())[0]
        config_model = platforms.get(platform, None)
        if config_model is None:
            raise ValueError(
                "Output config format not recognized, please see documentation for examples"
            )
        params = output_config[platform].copy()
        # validation on object creation
        config_obj = config_model(**params)

        params = {k: v for k, v in config_obj.dict().items() if v is not None}

        # convert credential files to b64 encoding
        if platform == "google_cloud_storage":
            if "credentials_id" not in params:
                location = params["service_credentials"]

                try:
                    with open(location, "rb") as binary_file:
                        encoded = base64.b64encode(binary_file.read())
                        params["service_credentials"] = encoded.decode("ascii")
                except (FileNotFoundError, OSError):
                    try:
                        creds = base64.b64decode(location)
                        assert "type" in json.loads(creds)
                    except:
                        raise ValueError(
                            "GCS service credentials should be either a Base64-encoded string"
                            + " of the JSON credentials file contents or a path to the JSON credentials file"
                        )
            else:
                if "service_credentials" in params:
                    c_id = params["credentials_id"]
                    warnings.warn(
                        "Both a stored credential ID and service credential argument were provided. "
                        + f'The store credential ID "{c_id}" will be used'
                    )
                    del params["service_credentials"]

        if platform != "destination":
            return {platform: params}
        else:
            return params

    def add_email_notification(self, address):
        """Add an email notification to the order

        Parameters
        ----------
        address : str
            Email address to receive order notifications"""
        self.request.notifications.append(EmailNotification(address=address))

    def add_sns_notification(self, topic_arn):
        """Add an AWS SNS notification topic to receive order notifications

        Parameters
        ----------
        topic_arn : str
            AWS SNS topic ARN to recieve order notifications"""
        self.request.notifications.append(SNSNotification(topic_arn=topic_arn))

    def __repr__(self):
        try:
            return f"<Order {self.order_id} ({self.status})>"
        except NotSubmitted:
            return "<Order (Not submitted)>"

    @property
    def submitted(self):
        if self.response:
            return self.response.id is not None
        else:
            return False

    @property
    @Submitted.required
    def finished(self):
        """The Order finished processing but may have failed"""

        return self.status != "RUNNING"

    @property
    @Submitted.required
    def running(self):
        """The Order is running"""

        return self.status == "RUNNING"

    @property
    @Submitted.required
    def succeeded(self):
        """The Order has finished running and has succeeded"""

        return self.status == "SUCCEEDED"

    @property
    @Submitted.required
    def usage(self):
        if self.response.order is None:
            # just submitted, refresh the order status
            response = self.get_order(self.order_id)
            self.response = OrderResponse(**response)
        return self.response.order.usage

    @property
    @Submitted.required
    def failed(self):
        """The Order has finished running but failed"""

        return self.status in ["FAILED", "ERROR"]

    @property
    @Submitted.required
    def order_id(self):
        """The Order ID"""

        return self.response.id

    @property
    @Submitted.required
    def status(self):
        """State of the order process: 'RUNNING', 'SUCCEEDED, or 'FAILED'"""

        if self.response.status == "RUNNING":
            response = self.get_order(self.order_id)
            self.response = OrderResponse(**response)
        return self.response.status

    @property
    def state(self):
        """Legacy version of `status`, will be deprecated in the future"""
        return self.status

    @classmethod
    def from_id(cls, order_id, session=None):
        """Create a Order object from an ID

        Parameters
        ----------
        order_id: str
            Order ID to hydrate into a Order object
        session : Session object, optional
            Authenticated session, such as from get_client_session()

        Returns
        -------
        Order"""

        response = cls.get_order(order_id, session)
        instance = cls()
        instance.response = OrderResponse(**response)
        return instance

    @classmethod
    def list_orders(
        cls,
        limit=10,
        starting_after="",
        ending_before="",
        start_date="",
        end_date="",
        filter=None,
        session=None,
    ):
        """Fetch user's orders

        Parameters
        ----------
        limit: int, default 10
            maximum number of orders to fetch
        starting_after: str
            the order_id after which further responses will be returned, paging forward
        ending_before: str
            the order_id before which further responses will be returned, paging backwards
        start_date: str
            starting date to filter, ISO-8601 YYYY-MM-DD
        start_date: str
            ending date to filter, ISO-8601 YYYY-MM-DD
        filter: str
            filter results that match values contained in the given key separated by a colon.
            Example: 'metadata.downstream_customer_id:abdc-534-b4dc47'
        session: Session object, optional
            Authenticated session, such as from get_client_session()

        Returns
        -------
        list
            Order objects matching parameters"""

        session = session or get_user_session()

        params = {
            "limit": limit,
            "starting_after": starting_after,
            "ending_before": ending_before,
            "start_date": start_date,
            "end_date": end_date,
        }

        if filter is not None:
            params["filter"] = filter

        key = lambda order: order["id"]
        responses = paginated_response(session, ard_url("order"), key, **params)

        return hydrate_with_responses(Order, OrderResponse, responses, session=session)

    @classmethod
    def get_order(cls, order_id, session=None):
        """Fetch raw data about an Order from an ID

        Parameters
        ----------
        order_id : str
            Order ID to fetch metadata for
        session: Session object, optional
            Authenticated session, such as from get_client_session()

        Returns
        -------
        dict
            API data for the given Order"""

        if not session:
            session = get_user_session()

        r = session.get(ard_url("order", "status", order_id))
        return r.json()

    @classmethod
    def send_order(cls, payload, session=None):
        """Send a request to the Order API

        Parameters
        ----------
        payload : dict
            Order API request payload
        session : Session object, optional
            Authenticated session, such as from get_client_session()

        Returns
        -------
        dict
            API response data for the given Order"""

        if not session:
            session = get_user_session()
        r = session.post(ard_url("order"), json=payload)
        return r.json()

    def submit(self):
        """Submit this Order to the API"""

        response = self.send_order(self.request.to_payload(), session=self.session)
        if self.submitted:
            warnings.warn("The Order has already been submitted")
            return
        if self.dry_run:
            self.response = OrderResponse(
                id="dry run",
                status="SUCCEEDED",
                order={"usage": response["usage"]},
                status_message="Order dry-run validation successful.",
            )
        else:
            self.response = OrderResponse(
                id=response["id"], status="RUNNING", status_message="Order submitted and running."
            )



"""Tools for working with Maxar ARD

Provides
--------
- Tools for interacting with the ARD API and files 
- SDK objects
- CLI tools

"""

from max_ard.admin import AccountManager
from max_ard.ard_collection import ARDCollection
from max_ard.order import Order
from max_ard.select import Select, SelectResult
from max_ard.monitor import Monitor

__all__ = ["ARDCollection", "Order", "Select", "SelectResult", "AccountManager"]
